/* -------------------------------------------------------------------------- */
/*                                                                            */
/* [sysbase1.c]                Standard Definitions ( Part 1 )                */
/*                                                                            */
/* Copyright (c) 1993 by D\olle, Manns                                        */
/* -------------------------------------------------------------------------- */

#include "sysbase0.h"

#include "mem_base.h"

/*I----------------- Byte/String & VarArg Handling -------------------------- */

/*INTERFACE*/
#define CORE            fprintf(STDOUT,"Objects = %ld\n", ObjCount())
#define BUG_CORE        bug1(ObjCount() == 0,"%ld objects",ObjCount())
#define BUG_CORE_EX(c)  bug1(ObjCount()+(c) == 0,"%ld objects",ObjCount()+(c))
#define New(TYPE)       (TYPE)NewMem(SizeOf(TYPE))

StdCPtr BytCopy(StdCPtr bytes, long len)
/* copies 'len' bytes of byte array 'bytes';
   allocs memory
*/
{ byte *res, *byt = (byte *)bytes; long  i;
  res = (byte*)NewMem(len);
  for(i = 0; i < len; i++)
    res[i] = byt[i];
  return (StdCPtr) res;
}

bool IsSpaceStr(string s)
/* whether string 's' is empty or 
   only contains space characters 
*/
{ 
  if( STR_EMPTY(s) ) return True;
  else
  {
    while( isspace(*s++) ) 
      ;
    return *s == '\0';
  }
}

size_t WCStrLen(wc_string ws)
/* number of wide characters in string 'ws'
   ( without the termination character )
*/
{ size_t len = 0, i = 0;
  while( ws[i] != WCHAR_LITERAL('\0') ) 
  {
    ++len; ++i;
  }
  return len;
}

wc_string SubWCStrCopy(wc_string str, long len)
/* copies 'len' wide characters of string 'str';
   appends 0, allocs memory
*/
{ wc_string res;
  res = (wc_string)BytCopy(str,(len+1)*sizeof(wc_char));
  res[len] = WCHAR_LITERAL('\0');
  return res;
}

wc_string WCStrCopy(wc_string Str)
/* copies wide character string 'Str'; allocs memory */
{
  return SubWCStrCopy(Str,WCStrLen(Str));
}

string WCToStrCopy(wc_string wc)
/* copies wide character string 'wc' into a string;
   without conversion, allocs memory 
*/
{ size_t len = WCStrLen(wc), i;
  string res = (string)NewMem(len+1);
  for( i=0; i < len; ++i )
    res[i] = (byte)wc[i];
  res[i] = '\0';
  return res;
}

wc_string WCStrToLower(wc_string st)
/* converts all letters in wide character string 'st' to lower case */
{ wc_string s;
#if defined(iswupper) && defined(towlower)
  for(s=st;*s;s++) if( iswupper(*s) ) *s = towlower(*s);
#else
  for(s=st;*s;s++) if( (wc_char)((byte)(*s)) == (*s) && isupper(*s) ) *s = tolower(*s);
#endif
  return st;
}

wc_string WCStrToUpper(wc_string st)
/* converts all letters in wide character string 'st' to upper case */
{ wc_string s;
#if defined(iswlower) && defined(towupper)
  for(s=st;*s;s++) if( iswlower(*s) ) *s = towupper(*s);
#else
  for(s=st;*s;s++) if( (wc_char)((byte)(*s)) == (*s) && islower(*s) ) *s = toupper(*s);
#endif
  return st;
}

string SubStrCopy(string str, long len)
/* copies 'len' characters of string 'str';
   appends 0, allocs memory
*/
{ string res;
  res = (string)BytCopy(str,len+1);
  res[len] = '\0';
  return res;
}

string StrCopy(string Str)
/* copies string 'Str'; allocs memory */
{
  return SubStrCopy(Str,strlen(Str));
}

string StrToLower(string st)
/* converts all letters in string 'st' to lower case */
{ string s;
  for(s=st;*s;s++) if( isupper(*s) ) *s = tolower(*s);
  return st;
}

string StrToUpper(string st)
/* converts all letters in string 'st' to upper case */
{ string s;
  for(s=st;*s;s++) if( islower(*s) ) *s = toupper(*s);
  return st;
}

/*INTERFACE*/
#define VA_START(ap,v) ((ap) = ((va_list)&(v)))

StdCPtr* GetArgLst(va_list args) 
/* variable arguments from list 'args
   The first argument contains the argument number ( len(args)-1 ). 
*/
{ StdCPtr* arglst; int argcnt = va_arg(args,int), i;
  arglst = (StdCPtr*)NewMem((argcnt+1)*sizeof(StdCPtr));
  arglst[0] = (StdCPtr)argcnt;
  for( i=0; i < argcnt; ++i ) arglst[i+1] = va_arg(args,StdCPtr);
  return( arglst );
}

/*I-------------------------- System Error ---------------------------------- */

#if defined( _MSDOS ) || defined( _WIN32 )
#include <windows.h>

string GetSystemError(void)
/* system error message; allocs memory */
{ long    rc;
  StdCPtr buf = (StdCPtr)NULL;
  string  txt;
  rc = FormatMessage
       (
         FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
         NULL,
         GetLastError(),
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default Language
         (LPTSTR)&buf,
         0,
         NULL
       );
  if( rc == 0 ) txt = StrCopy("");
  else          txt = StrCopy((string)buf);
  if( buf != (StdCPtr)NULL ) LocalFree(buf);
  return txt;
}

#else

  string GetSystemError(void)
  { int    e = errno;
    string t = strerror(e);
    return StrCopy(t==(string)NULL?"":t);
  }

#endif

/*I-------------------------- Program Execution ----------------------------- */

/*DOC_INTERFACE*/
#define STD_CMD_NOWAIT     0
#define STD_CMD_WAIT_INIT  1
#define STD_CMD_WAIT       2

#if defined( _WIN32 )
#include <windows.h>

  int runprg(string cmd, int cmdflg) // Win32
  /* exec program */
  { STARTUPINFO         StartInfo;
    PROCESS_INFORMATION ProcessInfo;
    long                ExitVal = 0, tmp;
    GetStartupInfo(&StartInfo);
    if( CreateProcess
        (
          NULL,cmd,
          NULL,NULL,FALSE,CREATE_NEW_PROCESS_GROUP,NULL,NULL,
          &StartInfo,&ProcessInfo
        ) == FALSE )
      ExitVal = GetLastError();
    else
    if( cmdflg == STD_CMD_WAIT )
    {
      tmp = WaitForSingleObject(ProcessInfo.hProcess,INFINITE);
      if( tmp == WAIT_FAILED )
        ExitVal = GetLastError();
      else
      if( GetExitCodeProcess(ProcessInfo.hProcess,&ExitVal) == FALSE )
        ExitVal = GetLastError();
    }
    else
    if( cmdflg == STD_CMD_WAIT_INIT )
    {
      if( WaitForInputIdle(ProcessInfo.hProcess,INFINITE/*msec*/) != 0 )
        ExitVal = GetLastError();
    }
    return (int)ExitVal;
  }

#else
#if defined( _MSDOS )
#include <windows.h>

  int runprg(string cmd, int cmdflg) // Windows 3.x
  /* exec program */
  { 
    if( cmdflg != STD_CMD_WAIT )
      return WinExec(cmd,SW_SHOW);
    else
      return system(cmd);
  }

#else
#if defined( MSDOS ) // Dos-GNU

  int runprg(string cmd, int cmdflg)
  /* exec program */
  { 
    return system(cmd);
  }

#else // Unix, Linux

int runprg(string cmd, int cmdflg) 
/* executes program 'cmd' 
   cmdflg=STD_CMD_NOWAIT    --> asynchron, non-blocking
   cmdflg=STD_CMD_WAIT_INIT --> waits for initialisation
   cmdflg=STD_CMD_WAIT      --> synchron, blocking
*/
{ 
  if( cmdflg == STD_CMD_WAIT )
    return system(cmd);
  else
  { string x = NewMem(strlen(cmd)+2+1);
    strcpy(x,cmd);
    strcat(x," &");
    return system(x);
  }
}

#endif

#endif

#endif

/*I----------------- DLL & dynamic Function Call Support -------------------- */

/*DOC_INTERFACE*/
typedef StdCPtr (*PCFUN)();

StdCPtr apply_fun_portable(PCFUN f, int cnt, StdCPtr* args)
/* portable dynamic call of C-function 'f' 
   with 'cnt' parameters in 'args'
   ( upto 9 parameters )
*/
{
  switch (cnt)
  {
    case 0: return f();
    case 1: return f(args[0]);
    case 2: return f(args[0],args[1]);
    case 3: return f(args[0],args[1],args[2]);
    case 4: return f(args[0],args[1],args[2],args[3]);
    case 5: return f(args[0],args[1],args[2],args[3],args[4]);
    case 6: return f(args[0],args[1],args[2],args[3],args[4],args[5]);
    case 7: return f(args[0],args[1],args[2],args[3],args[4],args[5],args[6]);
    case 8: return f(args[0],args[1],args[2],args[3],args[4],args[5],args[6],args[7]);
    case 9: return f(args[0],args[1],args[2],args[3],args[4],args[5],args[6],args[7],args[8]);
    default: 
      BUG; return NULL;
  }
}

#if defined( _MSDOS ) || defined( _WIN32 )
#define PUSH_FUN_ARG(arg) __asm mov eax, arg \
                          __asm push eax 
#define CALL_FUN(f)       __asm call f
         
StdCPtr apply_fun(PCFUN f, int cnt, StdCPtr* args)
/* dynamic call of C-function 'f' 
   with 'cnt' parameters in 'args' 
   ( restricted to intel )
*/
{ int i; StdCPtr arg;
  for( i=cnt-1; i >= 0; --i )
  {
    arg = args[i];
    PUSH_FUN_ARG(arg);
  }
  /*return*/ CALL_FUN(f); /* note that return is implied here. */
}

#else
#if !defined( MSDOS ) && defined ( __GNUC__ ) && #cpu (i386)
#define PUSH_FUN_ARG(arg) __asm__ volatile ( "pushl %0" : : "r" (arg) )
#define CALL_FUN(f)       __asm__ volatile ( "call *%0" : : "r" (f) )

  StdCPtr apply_fun(PCFUN f, int cnt, StdCPtr* args)
  { int i; StdCPtr arg;
    for( i=cnt-1; i >= 0; --i )
    {
      arg = args[i];
      PUSH_FUN_ARG(arg);
    }
    return (*f)(/*args*/);
    /*return CALL_FUN(f); note that return is implied here. */
  }

#else
/* non-intel-cpu || DOS-GNU-CC */
#define apply_fun apply_fun_portable

#endif

#endif

AbstractType( DL_Hdl ); /* Abstract handle type */

/*DOC_INTERFACE*/
#define DL_LOAD_LAZY   0

/*DOC_INTERFACE*/
#if !defined( _WIN32 ) && !defined( MSDOS ) && ( !defined( HAVE_CONFIG_H ) || defined( HAVE_LIBDL ) )
#include <dlfcn.h>
#define DL_LAZY         RTLD_LAZY
#define DL_NOW          RTLD_NOW
#define DL_BINDING_MASK RTLD_BINDING_MASK
#define DL_GLOBAL       RTLD_GLOBAL
#define DL_LOCAL        RTLD_LOCAL
#else
#include <windows.h>
#define DL_LAZY         0
#define DL_NOW          0
#define DL_BINDING_MASK 0
#define DL_GLOBAL       0
#define DL_LOCAL        0
#endif

#if defined( _MSDOS ) || defined( _WIN32 )
#include <windows.h>

  DL_Hdl DL_open(string dlname, int flag)
  { HMODULE dll = LoadLibrary((LPCTSTR)dlname);
    return (DL_Hdl)dll;
  }

  int DL_close(DL_Hdl dlhdl)
  { 
    if( FreeLibrary((HMODULE)dlhdl) == TRUE ) return 0;
    else return 1;
  }

  StdCPtr DL_symbol(DL_Hdl dlhdl, string symname)
  /* ReturnType: FARPROC */
  { FARPROC res = GetProcAddress((HMODULE)dlhdl,(LPCTSTR)symname);
    return (StdCPtr)res;
  }

  string DL_error(void)
  {
    return GetSystemError();
  }

#else
#if !defined( MSDOS ) && ( !defined( HAVE_CONFIG_H ) || defined( HAVE_LIBDL ) )

DL_Hdl DL_open(string dlname, int flag)
/* loads dll 'dlname' in mode 'flag' ( ERROR: NULL )
   LINUX: flag=RTLD_LAZY|RTLD_NOW|-1=RTLD_NOW, MS: not used     
*/
{ DL_Hdl dl = (DL_Hdl)dlopen(dlname,flag<0?RTLD_NOW:flag);
  return dl;
}

int DL_close(DL_Hdl dlhdl)
/* detaches dll 'dlhdl' ( SUCCESS: 0, ERROR: > 0 ) */
{
  return dlclose(dlhdl);
}

StdCPtr DL_symbol(DL_Hdl dlhdl, string symname)
/* address of symbol 'symname' within dll 'dlhdl' ( ERROR: NULL )
   LINUX: function|data, MS: function
*/
{
  return dlsym(dlhdl,symname);
}

string DL_error(void)
/* dll load error; allocs memory */
{ string t = dlerror();
  return t?t:""; //StrCopy(t==(string)NULL?"":t);
}

#else
/* DOS-GNU-CC, no ELF */

  DL_Hdl DL_open(string dlname, int flag)
  { 
    return (DL_Hdl)NULL;
  }

  int DL_close(DL_Hdl dlhdl)
  {
    return 1;
  }

  StdCPtr DL_symbol(DL_Hdl dlhdl, string symname)
  {
    return NULL;
  }

  string DL_error(void)
  {
    return GetSystemError();
  }

#endif

#endif

/*I---------------------- Network Support ----------------------------------- */

#if defined( _MSDOS ) || defined( _WIN32 )
#include <windows.h>

string UserName(void)
/* login name; allocs memory */
{ char buf[1000]; long dmy=999;
  GetUserName(buf,&dmy);
  return StrCopy(StrToLower(buf));
}

#else
#if !defined( MSDOS )

  string UserName(void)
  { string u = getenv("LOGNAME");
    return StrCopy(u!=(string)NULL?u:"");
  }

#else
/* DOS-GNU-CC */

  string UserName(void)
  { string u = getenv("USER");
    return StrCopy(u!=(string)NULL?u:"");
  }

#endif

#endif
