/* ------------------------------------------------------------------------ */
/*                                                                          */
/* [reg_exp.c]           Regular Expression Evaluation                      */
/*                                                                          */
/* Copyright (c) 1999 by D\olle, Manns                                      */
/* ------------------------------------------------------------------------ */

#include "standard.h"

#include "symbols.h"
#include "scn_base.h"
#include "scn_imp.h"
#include "scn_gen.h"
#include "literal.h"
#include "otab.h"
#include "hmap.h"
#include "hset.h"
#include "sink.h"

#include "rexp_lim.h"


/*I ------------------------ Macros, Types & Globals -------------------- */

// type of regular expression evaluator
ExternalType(Reg_T) 
{
  Scn_T pRScn; // common Regular Expression Scanner 
  Scn_T pVScn; // special Regular Expression Scanner 
};


/*I ------------------------ Create & Delete ---------------------------- */

static RegSet_T REG_String(string s)
{ RegSet_T reg; int len; string tstring = LIT_styx_lit_str(&len,s);
  reg = RegSet_StringN(tstring,len);
  FreeMem(tstring); 
  return reg;
}

static RegSet_T REG_Cset(string s)
{ RegSet_T reg; int len; string tstring = LIT_styx_lit_str(&len,s);
  reg = RegSet_CsetN(tstring,len);
  FreeMem(tstring); return reg;
}

static RegSet_T REG_self(RegSet_T pRegSet)
{
  return pRegSet;
}

static RegSet_T REG_plus(RegSet_T pRegSet)
{ RegSet_T pRes = RegSet_Plus(pRegSet);
  RegSet_Free(pRegSet);
  return pRes;
}

static RegSet_T REG_option(RegSet_T pRegSet)
{ RegSet_T pRes = RegSet_Option(pRegSet);
  RegSet_Free(pRegSet);
  return pRes;
}

static RegSet_T REG_star(RegSet_T pRegSet)
{ RegSet_T pRes = RegSet_Star(pRegSet);
  RegSet_Free(pRegSet);
  return pRes;
}

static RegSet_T REG_concat(RegSet_T pLeft, RegSet_T pRight)
{
  if( pLeft == (RegSet_T)NULL ) return pRight;
  else
  if( pRight == (RegSet_T)NULL ) return pLeft;
  else
  { RegSet_T pRes = RegSet_Concat(pLeft,pRight);
    RegSet_Free(pLeft); RegSet_Free(pRight);
    return pRes;
  }
}

static RegSet_T REG_diff(RegSet_T pLeft, RegSet_T pRight)
{ RegSet_T pRes = RegSet_Difference(pLeft,pRight);
  RegSet_Free(pLeft); RegSet_Free(pRight);
  return pRes;
}

static RegSet_T REG_union(RegSet_T pLeft, RegSet_T pRight)
{ RegSet_T pRes = RegSet_Union(pLeft,pRight);
  RegSet_Free(pLeft); RegSet_Free(pRight);
  return pRes;
}

static RegSet_T REG_range(RegSet_T pLeft, RegSet_T pRight)
{ RegSet_T pRes = (RegSet_T)NULL;
  short    al, bu, x;
  if(RegSet_isChar(pLeft) && RegSet_isChar(pRight))
  {
    al = RegSet_charVal(pLeft);
    bu = RegSet_charVal(pRight);
    if (al > bu)
    { 
      x = al; al = bu; bu = x; 
    }
    pRes = RegSet_Range(al,bu);
  }
  RegSet_Free(pLeft); RegSet_Free(pRight);
  return pRes;
}

static RegSet_T REG_sigma(void)
/* FIXME: charset size */
{
  return RegSet_Range(0,SCN_CHAR_SET_SIZE-1);
}

static void REG_freeExps(ROW(RegSet_T) pRegRow)
{ long lCnt = OT_CNT(pRegRow), lI;
  for( lI=0; lI < lCnt; ++lI )
    RegSet_Free(OT_GET(RegSet_T,pRegRow,lI));
  OT_CLEAR(pRegRow);
}

static RegSet_T REG_consExps(ROW(RegSet_T) pRegRow)
{ RegSet_T pRes = (RegSet_T)NULL, pRegSet;
  long     lCnt = OT_CNT(pRegRow), lI;
  if( lCnt == 0 ) pRes = RegSet_Empty();
  else
    for( lI=0; lI < lCnt; ++lI )
    {
      pRegSet = OT_GET(RegSet_T,pRegRow,lI);
      if( lI > 0 && pRes == (RegSet_T)NULL ) 
        RegSet_Free(pRegSet);
      else pRes = REG_concat(pRes,pRegSet);
    }
  OT_CLEAR(pRegRow);
  return pRes;
}

static RegSet_T REG_Exp(Scn_Stream pStream, int cEnd)
{ ROW(RegSet_T) pRegRow = OT_CREATE_ADT(RegSet_T);
  RegSet_T      (*pBRegFun)(RegSet_T pL, RegSet_T pR) = NULL;
  RegSet_T      (*pURegFun)(RegSet_T pRegSet) = NULL;
  RegSet_T      pRes = (RegSet_T)NULL, pRegSet1, pRegSet2;
  long          lRng = 0;
  int           cSym = 0;
  for(Stream_next(pStream); Stream_ctid(pStream) > 0; Stream_next(pStream))
  { c_string szTxt = symbolToString(Stream_csym(pStream));
//printf("TXT=%s\n",szTxt);
    if (Stream_ctid(pStream) == 1) 
    { 
      if( szTxt[0] == '+' )
      {
        if( OT_CNT(pRegRow) == 0 ) break;
        pRegSet1 = OT_TOP(RegSet_T,pRegRow); OT_POP(pRegRow); 
        pRegSet1 = REG_plus(pRegSet1);
        if( pRegSet1 == (RegSet_T)NULL ) break;
        OT_PUSH(pRegSet1,pRegRow);
      }
      else
      if( szTxt[0] == '.' && szTxt[1] == '.' ) 
      {
        if( OT_CNT(pRegRow) == 0 ) break;
        lRng = OT_CNT(pRegRow)+1;
      }
      else
      if( szTxt[0] == '|' || szTxt[0] == '-' )
      {
        if( OT_CNT(pRegRow) == 0 ) break;
        pBRegFun = szTxt[0] == '|' ? REG_union : REG_diff;
        pRegSet1 = REG_consExps(pRegRow);
        if( pRegSet1 == (RegSet_T)NULL ) break;
        pRegSet2 = REG_Exp(pStream,cEnd);
        if( pRegSet2 == (RegSet_T)NULL ) break;
        pRegSet1 = (*pBRegFun)(pRegSet1,pRegSet2);
        if( pRegSet1 == (RegSet_T)NULL ) break;
        OT_PUSH(pRegSet1,pRegRow);
        if( cEnd >= 0 ) { cSym = cEnd; break; }
      }
      else
      if( szTxt[0] == '.' ) 
      {
        pRegSet1 = REG_sigma();
        if( pRegSet1 == (RegSet_T)NULL ) break;
        OT_PUSH(pRegSet1,pRegRow);
      }
      else
      {
        if( szTxt[0] == '(' ) pURegFun = REG_self;
        else
        if( szTxt[0] == '{' ) pURegFun = REG_star;
        else
        if( szTxt[0] == '[' ) pURegFun = REG_option;
        else
        if( szTxt[0] == ')' || szTxt[0] == '}' || szTxt[0] == ']' )
        {
          cSym = (int)(szTxt[0]==')'?'(':szTxt[0]=='}'?'{':'['); break;
        }
        else BUG;
        pRegSet1 = REG_Exp(pStream, (int)szTxt[0]);
        if( pRegSet1 == (RegSet_T)NULL ) break;
        pRegSet1 = (*pURegFun)(pRegSet1);
        if( pRegSet1 == (RegSet_T)NULL ) break;
        OT_PUSH(pRegSet1,pRegRow);
      }
    }
    else
    { 
      pRegSet1 = ( Stream_ctid(pStream) == 2 )
                 ? REG_Cset(szTxt) : REG_String(szTxt);
      if( pRegSet1 == (RegSet_T)NULL ) break;
      OT_PUSH(pRegSet1,pRegRow);
    }
    if( lRng == OT_CNT(pRegRow) )
    {
      pRegSet1 = OT_TOP(RegSet_T,pRegRow); OT_POP(pRegRow);
      pRegSet2 = OT_TOP(RegSet_T,pRegRow); OT_POP(pRegRow);
      pRegSet1 = REG_range(pRegSet1,pRegSet2);
      if( pRegSet1 == (RegSet_T)NULL ) break;
      OT_PUSH(pRegSet1,pRegRow);
      lRng = 0;
    }
  }
  if( Stream_ctid(pStream) == -1 ) cSym = -1;
//printf("RNG=%ld,SYM=%ld,END=%ld\n",lRng,cSym,cEnd);
  if( lRng > 0 || cSym != cEnd ) REG_freeExps(pRegRow);
  else pRes = REG_consExps(pRegRow);
  OT_DEL_T(pRegRow);
  return pRes;
}

static Scn_T REG_Scanner(Scn_T pRScn, string szRegExp)
{ Scn_Stream pStream;
  RegSet_T   pRegSet;
  Scn_T      pVScn = (Scn_T)NULL;
  pStream = Stream_string(pRScn, szRegExp);
  Stream_defEofId(pStream, -1);
  Stream_defErrId(pStream,  0);
  Stream_defTokId(pStream, "Opr", 1);
  Stream_defTokId(pStream, "Set", 2);
  Stream_defTokId(pStream, "Seq", 3);
  pRegSet = REG_Exp(pStream, -1);
  Stream_close(pStream);
  Stream_free(pStream);
  if( pRegSet != (RegSet_T)NULL )
  {
    Scn_dfnBegin("REG");
    Scn_dfnToken("exp",(byte)0,pRegSet);
    pVScn = Scn_dfnEnd(False);
    RegSet_Free(pRegSet);
  }
  return pVScn;
}

void REG_drop(Reg_T pReg)
/* drop regular expression evaluator */
{
  BUG_NULL(pReg);
  Scn_free(pReg->pRScn);
  if( pReg->pVScn != (Scn_T)NULL ) Scn_free(pReg->pVScn);
  FreeMem(pReg);
}

Reg_T REG_create(c_string szRegExp)
/* create regular expression evaluator from 'szRegExp' */
{ Reg_T pReg = New(Reg_T);
  Scn_get_rexp(&pReg->pRScn);
  pReg->pVScn = REG_Scanner(pReg->pRScn, szRegExp);
  if( pReg->pVScn == (Scn_T)NULL )
  {
    REG_drop(pReg); return (Reg_T)NULL;
  }
  return pReg;
}

c_bool REG_recreate(Reg_T pReg, c_string szRegExp)
/* create regular expression evaluator from 'szRegExp' */
{ 
  BUG_NULL(pReg);
  if( pReg->pVScn != (Scn_T)NULL )
    Scn_free(pReg->pVScn);
  pReg->pVScn = REG_Scanner(pReg->pRScn, szRegExp);
  return( pReg->pVScn != (Scn_T)NULL );
}

/*I ------------------------ Evaluation --------------------------------- */

c_string REG_match(Reg_T pReg, c_string szTxt, long* lPatLen)
/* apply 'pReg' on 'szTxt'
   RC = start postion in 'szTxt' or NULL 
   lPatLen != NULL --> length of first matched value
*/
{ c_string szRes = (c_string)NULL, szTok;
  long     lIdx  = 0;
  BUG_NULL(pReg);
  if( pReg->pVScn != (Scn_T)NULL )
  { Scn_Stream pStream;
    pStream = Stream_string(pReg->pVScn, szTxt);
    Stream_concat_other(pStream);
    Stream_defEofId(pStream, -1);
    Stream_defErrId(pStream,  0);
    Stream_defTokId(pStream, "exp", 1);
    for(Stream_next(pStream); Stream_ctid(pStream) >= 0; Stream_next(pStream))
    {
      szTok = symbolToString(Stream_csym(pStream));
      if (Stream_ctid(pStream) == 1) 
      {
        if( lPatLen != (long*)NULL ) *lPatLen = strlen(szTok);
        szRes = szTxt + lIdx;
        break;
      }
      else lIdx += strlen(szTok);
    }
    Stream_close(pStream);
    Stream_free(pStream);
  }
  return szRes;
}

