/* ------------------------------------------------------------------------ */
/*                                                                          */
/* [scn_gen.c]                scanner generator                             */
/*                                                                          */
/* Copyright (c) 1993 by D\olle, Manns                                      */
/* ------------------------------------------------------------------------ */

/* TODO                                                        */
/* -  Write a document on the source                           */

#include "scn_io.h"

#include "list.h"
#include "ctx_app.h"
#include "scn_imp.h"

/*I------------------------- Type of regular sets ------------------------- */

AbstractType(RegSet_T); /* Regular Set */

/* -- Graph for finite automation -- */

#define RegSet_IMP(X) ((regset_t)(X))
#define RegSet_ADT(X) ((RegSet_T)(X))

ConcreteType(regset_t)
{
  int          id;
  List(faNode) Nodes; /* !empty; fst == start */
};

ConcreteType(faNode)
{
  int           id;
  regset_t      Belongs;
  List(faNode)  Source;  /* list of "source" nodes */
  int           term;    /* terminal state if > 0 */
  List(dfaEdge) Edges;
  int           flag1;
  int           flag2;
};

ConcreteType(dfaEdge)
{
  short  dfa_lower;
  short  dfa_upper;
  faNode dfa_toNode;
};

ConcreteType(nfaEdge)
{
  short        nfa_lower;
  short        nfa_upper;
  List(faNode) nfa_toNodes;
};

static void RegSet_Print(RegSet_T x); /* forward */

/* -- Edge constructor -- */

static dfaEdge newDfaEdge(short lower, short upper, faNode toNode)
{ dfaEdge res = New(dfaEdge);
  res->dfa_lower  = lower;
  res->dfa_upper  = upper;
  res->dfa_toNode = toNode;
  return res;
}

static nfaEdge newNfaEdge(short lower, short upper, List(faNode) toNodes)
{ nfaEdge res = New(nfaEdge);
  res->nfa_lower   = lower;
  res->nfa_upper   = upper;
  res->nfa_toNodes = toNodes;
  return res;
}

/*I----------------------------- freeing ---------------------------------- */

static void freeNothing(Abs_T dummy)
{
}

static void freeNfaEdge(nfaEdge x)
{
  freeList(x->nfa_toNodes, freeNothing);
  FreeMem(x);
}

static void freeDfaNode(faNode x)
/* frees dfa node */
{
  freeList(x->Source, freeNothing);
  freeList(x->Edges, FreeMem);
  FreeMem(x);
}

void RegSet_Free(RegSet_T g)
/* frees 'g' */
{
  freeList(RegSet_IMP(g)->Nodes, (void (*)(Abs_T x)) freeDfaNode);
  FreeMem(g);
}

/*I-------------------------- RegSet constructors ------------------------- */

/*INTERFACE*/
/* Hinweis zur Speicherverwaltung. Alle Funktionen, die einen 'RegSet_T' */
/* als Ergebnis haben, erzeugen diesen neu. Seine Freigabe ist           */
/* Verpflichtung fr den Klienten                                        */

/* -- basic graph constructors -- */

static int RegSet_TId = 1;

RegSet_T RegSet_Epsilon()
/* "" */
{ faNode s0   = New(faNode);
  regset_t res = New(regset_t);
  res->id     = RegSet_TId ++;
  res->Nodes  = cons(s0,NULL);
  s0->id      = 0;
  s0->Belongs = res;
  s0->Source  = NULL;
  s0->term    = True; /* terminal */
  s0->Edges   = NULL;
  s0->flag1   = 0;
  s0->flag2   = 0;
  return RegSet_ADT(res);
}

RegSet_T RegSet_Empty()
/* '' */
{ faNode s0    = New(faNode);
  regset_t res = New(regset_t);
  res->id      = RegSet_TId ++;
  res->Nodes   = cons(s0,NULL);
  s0->id       = 0;
  s0->Belongs  = res;
  s0->Source   = NULL;
  s0->term     = False; /* non - terminal */
  s0->Edges    = NULL;
  s0->flag1    = 0;
  s0->flag2    = 0;
  return RegSet_ADT(res);
}

RegSet_T RegSet_Range(short dfa_lower, short dfa_upper)
/* L .. U */
{ faNode s0   = New(faNode);
  faNode s1   = New(faNode);
  regset_t res = New(regset_t);
  bug(dfa_lower <= dfa_upper,
      "RegSet_Range(%d,%d) must be a range",dfa_lower,dfa_upper);
  res->id     = RegSet_TId ++;
  res->Nodes  = cons(s0,cons(s1,NULL));
  s0->id      = 0;
  s0->Belongs = res;
  s0->Source  = NULL;
  s0->term    = False; /* non - terminal */
  s0->Edges   = cons(newDfaEdge(dfa_lower,dfa_upper,s1),NULL);
  s0->flag1   = 0;
  s0->flag2   = 0;
  s1->id      = 1;
  s1->Belongs = res;
  s1->Source  = NULL;
  s1->term    = True; /* terminal */
  s1->Edges   = NULL;
  s1->flag1   = 0;
  s1->flag2   = 0;
  return RegSet_ADT(res);
}

RegSet_T RegSet_Char(short c)
/* 'c' */
{
  return RegSet_Range(c,c);
}

/* -- Graph Combinator -- */

static int lex(int A, int B) { return A ? A : B; }

static bool compNodes(faNode a, faNode b)
/* order Nodes */
{
  return lex( a->Belongs->id - b->Belongs->id, a->id - b->id );
}

static List(faNode) insertState(faNode x, List(faNode) y)
/* inserts a state into a source id */
/* we keep them ordered for identification */
{
  return empty(y)                        ? cons(x,y)
       : compNodes(x,fst(faNode,y)) == 0 ? y
       : compNodes(x,fst(faNode,y)) <  0 ? cons(x,y)
       : ((*rst_ref(y) = insertState(x,rst(y))), y);
}

static List(nfaEdge) copyEdges
                     (
                       List(dfaEdge) x, RegSet_T lambdaFrom, RegSet_T lambdaTo
                     )
/* copies an DFA-Edge to a NFA-Edge extending lambda edges */
{
  if (empty(x))
    return NULL;
  else
  { List(faNode) src = NULL;
    faNode nod = fst(dfaEdge,x)->dfa_toNode;
    src = insertState(nod,src);
    if (nod->Belongs == RegSet_IMP(lambdaFrom) && nod->term > 0)
      src = insertState(fst(faNode,RegSet_IMP(lambdaTo)->Nodes),src);
    return cons(newNfaEdge(fst(dfaEdge,x)->dfa_lower,
                        fst(dfaEdge,x)->dfa_upper,
                        src ),
                copyEdges(rst(x), lambdaFrom,lambdaTo) );
  }
}

static List(faNode) nodesUnion(List(faNode) x, List(faNode) y)
{
  if ( empty(x) &&  empty(y))
    return NULL;
  if ( empty(x) && !empty(y))
    return cons(fst(faNode,y),nodesUnion(NULL,rst(y)));
  if (!empty(x) &&  empty(y))
    return cons(fst(faNode,x),nodesUnion(rst(x),NULL));
  if (compNodes(fst(faNode,x),fst(faNode,y)) < 0)
    return cons(fst(faNode,x),nodesUnion(rst(x),y));
  else
  if (compNodes(fst(faNode,x),fst(faNode,y)) > 0)
    return cons(fst(faNode,y),nodesUnion(x,rst(y)));
  else
    return cons(fst(faNode,x),nodesUnion(rst(x),rst(y)));
}

static bool equalStates(List(faNode) x, List(faNode) y)
{
  for (; !empty(x) && !empty(y); x=rst(x), y=rst(y))
    if (fst(faNode,x) != fst(faNode,y))
      return False;
  return empty(x) && empty(y);
}

static List(nfaEdge) consMerge(nfaEdge x, List(nfaEdge) y)
/* joins two edges if possible */
{
  if (!empty(y) &&
      x->nfa_upper+1 == fst(nfaEdge,y)->nfa_lower &&
      equalStates(x->nfa_toNodes,fst(nfaEdge,y)->nfa_toNodes) )
  {
    fst(nfaEdge,y)->nfa_lower = x->nfa_lower;
    freeNfaEdge(x);
    return y;
  }
  else
    return cons(x,y);
}

static List(nfaEdge) mergeEdges(List(nfaEdge) x, List(nfaEdge) y)
/* merges NFA-Edges x with NFA-Edges y */
/* Note that both lists are freed */
{ nfaEdge xe, ye; List(nfaEdge) xy, res;
  if (empty(x)) return y;
  if (empty(y)) return x;
  xy = x; xe = fst(nfaEdge,x); x = rst(x); FreeMem(xy);
  xy = y; ye = fst(nfaEdge,y); y = rst(y); FreeMem(xy);
  if (xe->nfa_upper > ye->nfa_upper)
  {
    { nfaEdge       a = xe, b = ye; xe = b; ye = a; }
    { List(nfaEdge) a = x,  b = y;  x  = b; y  = a; }
  }
  if (xe->nfa_upper < ye->nfa_lower)
    return consMerge(xe,mergeEdges(x,cons(ye,y)));
  /* we have a true intersection here */
  res = consMerge
        (
          newNfaEdge(MAX(xe->nfa_lower,ye->nfa_lower),
                  xe->nfa_upper,
                  nodesUnion(xe->nfa_toNodes, ye->nfa_toNodes) ),
          mergeEdges(x,
                     xe->nfa_upper < ye->nfa_upper
                     ? cons(newNfaEdge(xe->nfa_upper+1,ye->nfa_upper,
                                       nodesUnion(ye->nfa_toNodes, NULL)
                                   ),y )
                     : y )
        );
  if (xe->nfa_lower > ye->nfa_lower)
  { nfaEdge a = xe, b = ye; xe = b; ye = a; }
  if (xe->nfa_lower < ye->nfa_lower)
    res = consMerge(newNfaEdge(xe->nfa_lower, ye->nfa_lower-1,
                               nodesUnion(xe->nfa_toNodes,NULL) ), res );
  freeNfaEdge(xe); freeNfaEdge(ye);
  return res;
}

/* -- */

static faNode identifySource(RegSet_T g, List(faNode) x)
/* identify state in g with Source equal to x */
{ List(faNode) *t; int i; faNode s0;
  for(i = 0, t = &RegSet_IMP(g)->Nodes; !empty(*t); i += 1, t = rst_ref(*t))
    if (equalStates( fst(faNode,*t)->Source, x))
    { freeList(x, freeNothing); return fst(faNode,*t); }
  s0 = New(faNode);
  s0->id      = i;
  s0->Belongs = RegSet_IMP(g);
  s0->Source  = x;
  s0->term    = False;
  s0->Edges   = NULL;
  s0->flag1   = 0;
  s0->flag2   = 0;
  *t = cons(s0,NULL);
  return s0;
}

static List(faNode) NfaToDfaEdges(RegSet_T g, List(nfaNode) x)
/* translates NfaEdges to DdfaEdges identifying or adding states to g */
{ List(nfaNode) t;
  for(t = x; !empty(t); t = rst(t))
  { nfaEdge e = fst(nfaEdge,t);
    ((dfaEdge)e)->dfa_toNode = identifySource(g, e->nfa_toNodes);
  }
  return (List(faNode)) x;
}

static List(faNode) RegSet_Starts
                    (
                      List(RegSet_T) graphs,
                      RegSet_T lambdaFrom,
                      RegSet_T lambdaTo
                    )
/* returns the list of start nodes of graphs */
/* frees x */
{ List(RegSet_T) x; List(faNode) y; faNode n;
  if (empty(graphs)) return NULL;
  x = rst(graphs);
  n = fst(faNode,fst(regset_t,graphs)->Nodes);
  FreeMem(graphs);
  y = RegSet_Starts(x, lambdaFrom,lambdaTo);
  if (n->Belongs == lambdaFrom && n->term > 0)
    y = insertState(fst(faNode,RegSet_IMP(lambdaTo)->Nodes),y);
  return insertState(n,y);
}

static RegSet_T RegSet_Merge
                (
                  List(RegSet_T) graphs, RegSet_T lambdaFrom, RegSet_T lambdaTo
                )
{ faNode s0   = New(faNode);
  List(faNode) Todo;
  regset_t res = New(regset_t);
  res->id     = RegSet_TId ++;
  res->Nodes  = cons(s0,NULL);
  s0->id      = 0;
  s0->Belongs = res;
  s0->Source  = RegSet_Starts(graphs,lambdaFrom,lambdaTo);
  s0->term    = False;
  s0->Edges   = NULL;
  s0->flag1   = 0;
  s0->flag2   = 0;
  for (Todo = res->Nodes; !empty(Todo); Todo = rst(Todo))
  { List(faNode) Nds; List(nfaEdge) NfaEdges = NULL;
    for( Nds = fst(faNode,Todo)->Source; !empty(Nds); Nds = rst(Nds))
      NfaEdges = mergeEdges
                 (
                   NfaEdges,
                   copyEdges(fst(faNode,Nds)->Edges,lambdaFrom,lambdaTo)
                 );
    fst(faNode,Todo)->Edges = NfaToDfaEdges(res, NfaEdges);
  }
  return RegSet_ADT(res);
}

/* -- graph reduction -- */

/* Ziel:    von allen Zustaenden aus kann ein terminaler Zustand erreicht */
/*          werden.                                                       */
/* Methode: entferne alle anderen                                         */

#define reachesNxt flag1
#define reachesTrm flag2

static void reachNodes(regset_t reg)
/* build the closure of terminal nodes via egdes */
{ List(faNode) nodes; List(dfaEdges) edges; bool changed = True;
  while (changed)
  { changed = False;
    for (nodes = reg->Nodes; !empty(nodes); nodes = rst(nodes))
    { faNode nod = fst(faNode,nodes);
      for (edges = nod->Edges; !empty(edges); edges = rst(edges))
        nod->reachesNxt |= fst(dfaEdge,edges)->dfa_toNode->reachesTrm;
    }
    for (nodes = reg->Nodes; !empty(nodes); nodes = rst(nodes))
    { faNode nod = fst(faNode,nodes);
      changed |= (nod->reachesNxt != nod->reachesTrm);
      nod->reachesTrm |= nod->reachesNxt;
    }
  }
}

static List(dfaEdge) combineRanges(List(dfaEdge) edges)
/* combines ranges if possible */
{ dfaEdge e1, e2; List(dfaEdge) rest;
  if (empty(edges) || empty(rst(edges))) return edges;
  rest = combineRanges(rst(edges));
  *rst_ref(edges) = rest;
  e1 = fst(dfaEdge,edges);
  e2 = fst(dfaEdge,rst(edges));
  if (e1->dfa_upper+1 != e2->dfa_lower || e1->dfa_toNode != e2->dfa_toNode)
    return edges;
  /* combine */
  e2->dfa_lower = e1->dfa_lower;
  FreeMem(e1); FreeMem(edges);
  return rest;
}

static void reduceGraph(RegSet_T g)
/* reduce graph g */
{ List(faNode) nods, *nodsp; int i;
  /* initialize marks */
  for (nods = RegSet_IMP(g)->Nodes; !empty(nods); nods = rst(nods))
  { faNode nod = fst(faNode,nods);
    nod->reachesTrm = (nod->term > 0);
    nod->reachesNxt = (nod->term > 0);
  }
  /* mark */
  reachNodes(RegSet_IMP(g));
  fst(faNode,RegSet_IMP(g)->Nodes)->reachesTrm = True;
  /* remove all edges leading to blind states */
  for (nods = RegSet_IMP(g)->Nodes; !empty(nods); nods = rst(nods))
  { faNode nod = fst(faNode,nods); List(dfaEdge) *edges;
    for (edges = &nod->Edges; !empty(*edges); )
    { dfaEdge edge = fst(dfaEdge,*edges);
      if (!edge->dfa_toNode->reachesTrm)
      { List(dfaEdge) here = *edges;
        *edges = rst(*edges); FreeMem(edge); FreeMem(here);
      }
      else
        edges = rst_ref(*edges);
    }
    nod->Edges = combineRanges(nod->Edges);
  }
  /* remove all blind states */
  for (i = 0, nodsp = &RegSet_IMP(g)->Nodes; !empty(*nodsp); )
  { faNode nod = fst(faNode,*nodsp);
    if (!nod->reachesTrm)
    { List(faNode) here = *nodsp; *nodsp = rst(*nodsp); FreeMem(here); FreeMem(nod); }
    else
    { nod->id = i; i += 1; nodsp = rst_ref(*nodsp); }
  }
}

/* -- graph minimation -- */

/* Ziel:    der Graph enthaelt keine aequivalenten Zustaende */
/* Methode: entferne alle aequivalenten Zustaende            */

#define currEquiv flag1
#define prevEquiv flag2

static faNode findNode(RegSet_T g, int NodeId)
{ List(faNode) nods;
  for (nods = RegSet_IMP(g)->Nodes; !empty(nods); nods = rst(nods))
    if ( fst(faNode,nods)->id == NodeId )
      return fst(faNode,nods);
  BUG; return NULL; /* dummy */
}

static bool sameEquiv(List(faNode) nods)
/* copy curr to prev; return true if all same */
{ bool res = True;
  for (; !empty(nods); nods = rst(nods))
  { faNode nod = fst(faNode,nods);
    res = res && (nod->prevEquiv == nod->currEquiv);
    nod->prevEquiv = nod->currEquiv;
  }
  return res;
}

static bool sameEquivClass(List(dfaEdge) edgs1, List(dfaEdge) edgs2)
/* return if both edge-lists are equivalent relative to prevEquiv */
{ dfaEdge edg1, edg2 ;
Loop1:
  if (empty(edgs1) && empty(edgs2)) return True;
  if (empty(edgs1) || empty(edgs2)) return False;
  edg1 = fst(dfaEdge,edgs1);
  edg2 = fst(dfaEdge,edgs2);
  if (edg1->dfa_lower != edg2->dfa_lower) return False;
Loop2:
  if (edg1->dfa_toNode->prevEquiv != edg2->dfa_toNode->prevEquiv) return False;
  if (edg1->dfa_upper == edg2->dfa_upper)
  {
    edgs1 = rst(edgs1); edgs2 = rst(edgs2);
    goto Loop1;
  }
  else
  if (edg1->dfa_upper < edg2->dfa_upper)
  {
    if (empty(rst(edgs1)) ||
        edg1->dfa_upper+1 != ((dfaEdge)rst(edgs1))->dfa_lower)
      return False;
    edgs1 = rst(edgs1); edg1 = fst(dfaEdge,edgs1);
    goto Loop2;
  }
  else
  {
    if (empty(rst(edgs2)) ||
        edg2->dfa_upper+1 != ((dfaEdge)rst(edgs2))->dfa_lower)
      return False;
    edgs2 = rst(edgs2); edg2 = fst(dfaEdge,edgs2);
    goto Loop2;
  }
}

static void minimizeGraph(RegSet_T g)
/* reduce graph g */
{ List(faNode) nods, nods1, nods2, *nodsp; int i;
  /* initialize equivalence */
  for (nods1 = RegSet_IMP(g)->Nodes; !empty(nods1); nods1 = rst(nods1))
  { faNode nod1 = fst(faNode,nods1), nod2;
    for (nods2 = RegSet_IMP(g)->Nodes; !empty(nods2); nods2 = rst(nods2))
    { nod2 = fst(faNode,nods2); if (nod1->term == nod2->term) break; }
    nod2 = fst(faNode,nods2);
    nod1->prevEquiv = -1; nod1->currEquiv = nod2->id;
  }
  while (!sameEquiv(RegSet_IMP(g)->Nodes))
  { /* next equivalence */
    for (nods1 = RegSet_IMP(g)->Nodes; !empty(nods1); nods1 = rst(nods1))
    { faNode nod1 = fst(faNode,nods1), nod2;
      for (nods2 = RegSet_IMP(g)->Nodes; !empty(nods2); nods2 = rst(nods2))
      { nod2 = fst(faNode,nods2);
        if (nod1->prevEquiv == nod2->prevEquiv &&
            sameEquivClass(nod1->Edges, nod2->Edges))
          break;
      }
      nod2 = fst(faNode,nods2);
      nod1->currEquiv = nod2->id;
    }
  }
  /* patch the edges                   */
  for (nods = RegSet_IMP(g)->Nodes; !empty(nods); nods = rst(nods))
  { faNode nod = fst(faNode,nods); List(dfaEdge) edges;
    for (edges = nod->Edges; !empty(edges); edges = rst(edges))
    { dfaEdge edge = fst(dfaEdge,edges);
      edge->dfa_toNode = findNode(g,edge->dfa_toNode->currEquiv);
    }
    nod->Edges = combineRanges(nod->Edges);
  }
  /* drop all equivalents beside first */
  for (i = 0, nodsp = &RegSet_IMP(g)->Nodes; !empty(*nodsp); )
  { faNode nod = fst(faNode,*nodsp);
    if (nod->currEquiv != nod->id)
    {
      List(faNode) here = *nodsp; *nodsp = rst(*nodsp);
      FreeMem(here); freeDfaNode(nod);
    }
    else
    { nod->id = i; i += 1; nodsp = rst_ref(*nodsp); }
  }
}

static bool hasTerminal(List(faNode) x, RegSet_T a)
/* whether a terminal state of 'a' is in source list 'x' */
{
  for (; !empty(x); x = rst(x))
  { faNode nod = fst(faNode,x);
    if (nod->Belongs == RegSet_IMP(a) && nod->term > 0)
      return True;
  }
  return False;
}

/* -- graph operations -- */

RegSet_T RegSet_Copy(RegSet_T a)
/* a */
{ RegSet_T res; List(faNode) lst;
  res = RegSet_Merge( cons(a,NULL), NULL,NULL );
  for (lst = RegSet_IMP(res)->Nodes; !empty(lst); lst = rst(lst))
  { faNode nod = fst(faNode,lst);
    nod->term = hasTerminal(nod->Source,a);
    freeList(nod->Source, freeNothing); nod->Source = NULL;
  }
  return res;
}

RegSet_T RegSet_Union(RegSet_T a, RegSet_T b)
/* a | b */
{ RegSet_T res; List(faNode) lst;
  if (a == b) return RegSet_Copy(a);
  res = RegSet_Merge( cons(a,cons(b, NULL)), NULL,NULL );
  for (lst = RegSet_IMP(res)->Nodes; !empty(lst); lst = rst(lst))
  { faNode nod = fst(faNode,lst);
    nod->term = hasTerminal(nod->Source,a) || hasTerminal(nod->Source,b);
    freeList(nod->Source, freeNothing); nod->Source = NULL;
  }
  reduceGraph(res); minimizeGraph(res);
  return res;
}

RegSet_T RegSet_Intersection(RegSet_T a, RegSet_T b)
/* a intersection b */
{ RegSet_T res; List(faNode) lst;
  if (a == b) return RegSet_Copy(a);
  res = RegSet_Merge( cons(a,cons(b, NULL)), NULL,NULL );
  for (lst = RegSet_IMP(res)->Nodes; !empty(lst); lst = rst(lst))
  { faNode nod = fst(faNode,lst);
    nod->term = hasTerminal(nod->Source,a) && hasTerminal(nod->Source,b);
    freeList(nod->Source, freeNothing); nod->Source = NULL;
  }
  reduceGraph(res); minimizeGraph(res);
  return res;
}

RegSet_T RegSet_Difference(RegSet_T a, RegSet_T b)
/* a - b */
{ RegSet_T res; List(faNode) lst;
  if (a == b) return RegSet_Empty();
  res = RegSet_Merge( cons(a,cons(b, NULL)), NULL,NULL );
  for (lst = RegSet_IMP(res)->Nodes; !empty(lst); lst = rst(lst))
  { faNode nod = fst(faNode,lst);
    nod->term = hasTerminal(nod->Source,a) && !hasTerminal(nod->Source,b);
    freeList(nod->Source, freeNothing); nod->Source = NULL;
  }
  reduceGraph(res); minimizeGraph(res);
  return res;
}

RegSet_T RegSet_Concat(RegSet_T a, RegSet_T b)
/* a b */
{ RegSet_T res,c; List(faNode) lst;
  c = (a == b) ? RegSet_Copy(a) : b;
  res = RegSet_Merge( cons(a,NULL), a,c );
  for (lst = RegSet_IMP(res)->Nodes; !empty(lst); lst = rst(lst))
  { faNode nod = fst(faNode,lst);
    nod->term = hasTerminal(nod->Source,b);
    freeList(nod->Source, freeNothing); nod->Source = NULL;
  }
  if (a==b) RegSet_Free(c);
  reduceGraph(res);
  minimizeGraph(res);
  return res;
}

RegSet_T RegSet_Star(RegSet_T a)
/* { a } */
{ RegSet_T res; List(faNode) lst;
  res = RegSet_Merge( cons(a,NULL), a,a );
  for (lst = RegSet_IMP(res)->Nodes; !empty(lst); lst = rst(lst))
  { faNode nod = fst(faNode,lst);
    nod->term = hasTerminal(nod->Source,a) ||
                fst(faNode,nod->Source) == fst(faNode,RegSet_IMP(a)->Nodes);
    freeList(nod->Source, freeNothing); nod->Source = NULL;
  }
  reduceGraph(res); minimizeGraph(res);
  return res;
}

/* -- combinations -- */

RegSet_T RegSet_CsetN(string s,int len)
/* '...'[0:len) */
{
  if (len > 0)
  { RegSet_T a = RegSet_Char(0xff & s[0]);
    RegSet_T b = RegSet_CsetN(s+1,len-1);
    RegSet_T c = RegSet_Union(a, b);
    RegSet_Free(a); RegSet_Free(b);
    return c;
  }
  else
    return RegSet_Empty();
}

RegSet_T RegSet_Cset(string s)
/* '...' */
{
  return RegSet_CsetN(s,strlen(s));
}

RegSet_T RegSet_StringN(string s,int len)
/* "..."[0:len) */
{
  if (len > 0)
  { RegSet_T a = RegSet_Char(0xff & s[0]);
    RegSet_T b = RegSet_StringN(s+1,len-1);
    RegSet_T c = RegSet_Concat(a, b);
    RegSet_Free(a); RegSet_Free(b);
    return c;
  }
  else
    return RegSet_Epsilon();
}

RegSet_T RegSet_String(string s)
/* "..." */
{
  return RegSet_StringN(s,strlen(s));
}

RegSet_T RegSet_Option(RegSet_T a)
/* [ a ] */
{ RegSet_T b = RegSet_Epsilon();
  RegSet_T c = RegSet_Union(a,b);
  RegSet_Free(b);
  return c;
}

RegSet_T RegSet_Plus(RegSet_T a)
/* a + */
{ RegSet_T b = RegSet_Star(a);
  RegSet_T c = RegSet_Concat(a,b);
  RegSet_Free(b);
  return c;
}

bool RegSet_isChar(RegSet_T a)
/* if a = 'x' */
{ faNode s0 = fst(faNode,RegSet_IMP(a)->Nodes);
  bool res  =  !empty(rst(RegSet_IMP(a)->Nodes))
            &&  empty(rst(rst(RegSet_IMP(a)->Nodes)))
            && !empty(s0->Edges)     && empty(rst(s0->Edges));
  dfaEdge e = res ? fst(dfaEdge,s0->Edges) : NULL;
  return res && e->dfa_lower == e->dfa_upper;
}

short RegSet_charVal(RegSet_T a)
/* x in a = 'x' */
{ int res;
  bug0( RegSet_isChar(a), "single character set requiered" );
  res = fst(dfaEdge,fst(faNode,RegSet_IMP(a)->Nodes)->Edges)->dfa_lower;
  return res;
}

/*I------------------- The Production Of The Scanner ---------------------- */

/* -- The Production Of The Scanner -- */

struct DefinedGraphs
{
  string  name;
  int     usage;
  RegSet_T value;
};

#define MAX_ENTRIES 50
static struct DefinedGraphs Table[MAX_ENTRIES];
static int    topEntry;
static bool   Scn_defining = False;
static string Scn_name;

void Scn_dfnBegin(string name)
/* begin a scanner definition */
{
  bug0(!Scn_defining,"allready defining a scanner");
  Scn_name = StrCopy(name);
  Scn_defining = True;
  topEntry = 1;
  Table[0].name  = SCN_TOK_Other;
  Table[0].usage = 0;
  Table[0].value = (RegSet_T)NULL;
/*
  topEntry = 2;
  Table[1].name  = StrCopy("[empty]");
  Table[1].usage = 0;
  Table[1].value = RegSet_Epsilon();
*/
}

void Scn_dfnToken(string name, byte tok_flags, RegSet_T value)
/* add a new token to the scanner under production */
{ int i;
  bug0(Scn_defining,"missing Scn_dfnBegin");
  for (i = 1; i < topEntry; i++)
    bug1(strcmp(Table[i].name,name), "duplicate name \'%s\'", name);
  bug0(topEntry < MAX_ENTRIES, "too many tokens defined");
  Table[topEntry].name  = StrCopy(name);
  Table[topEntry].usage = tok_flags;
  Table[topEntry].value = RegSet_Copy(value);
  topEntry += 1;
}

static bool ConflictFlag;

static void pConflicts(List(faNode) nodes)
{
  if (length(nodes) > 1)
  {
    ConflictFlag = True;
    fprintf(STDERR,"Conflict between ");
    for (;!empty(nodes); nodes = rst(nodes))
    { faNode node = fst(faNode,nodes);
      fprintf(STDERR,"'%s'",Table[node->Belongs->id].name);
      if (length(nodes) > 1)
        if (length(nodes) == 2)
          fprintf(STDERR," and ");
        else
          fprintf(STDERR,", ");
    }
    fprintf(STDERR,".\n");
  }
}

static List(faNode) terminalSource(List(faNode) nodes)
/* removes all non-terminal states from the source of a node */
{
  if(empty(nodes))
    return nodes;
  else
  { List(faNode) r = terminalSource(rst(nodes));
    if( fst(faNode,nodes)->term )
    { *rst_ref(nodes) = r; return nodes; }
    else
    { FreeMem(nodes); return r; }
  }
}

static void terminalScanner(RegSet_T graph)
/* removes all non-terminal states from the sources of a graph */
{ List(faNode) nodes = RegSet_IMP(graph)->Nodes;
  for (; !empty(nodes); nodes = rst(nodes))
  { faNode node = fst(faNode,nodes);
    node->Source = terminalSource(node->Source);
    node->term = empty(node->Source)
                 ? 0
                 : fst(faNode,node->Source)->Belongs->id;
    pConflicts(node->Source);
  }
}

/* new version of scanner table production */

static Scn_T ScnTable (RegSet_T scanner)
{ int i, states = length(RegSet_IMP(scanner)->Nodes);
  short *StaEdg;
  short *StaFin;
  byte  *EdgeC;
  short *EdgeS;
  List(faNode) nodes; List(dfaEdge) edges;
  scn_t res;
  int clow, edgcnt;
  /* count edges */
  edgcnt = 0;
  for (nodes = RegSet_IMP(scanner)->Nodes; !empty(nodes); nodes = rst(nodes))
  { faNode node = fst(faNode,nodes);
    clow = 0;
    for (edges = node->Edges; !empty(edges); edges = rst(edges))
    { dfaEdge edge = fst(dfaEdge,edges);
      edgcnt += (clow < edge->dfa_lower)+1;
      clow = edge->dfa_upper+1;
    }
    if (clow < SCN_CHAR_SET_SIZE) edgcnt += 1;
  }
  /* make edges and states */
  StaEdg = (short*)NewMem( (states+1) * sizeof(short) );
  StaFin = (short*)NewMem( states     * sizeof(short) );
  EdgeC  = (byte *)NewMem( edgcnt     * sizeof(byte ));
  EdgeS  = (short*)NewMem( edgcnt     * sizeof(short));
  edgcnt = 0;
  for (nodes = RegSet_IMP(scanner)->Nodes; !empty(nodes); nodes = rst(nodes))
  { faNode node = fst(faNode,nodes);
    StaEdg[node->id] = edgcnt;
    StaFin[node->id] = node->term;
    clow = 0;
    for (edges = node->Edges; !empty(edges); edges = rst(edges))
    { dfaEdge edge = fst(dfaEdge,edges);
      if (clow < edge->dfa_lower)
      {
        EdgeC[edgcnt] = clow; EdgeS[edgcnt] = 0; edgcnt += 1;
      }
      EdgeC[edgcnt] = (byte)edge->dfa_lower;
      EdgeS[edgcnt] = edge->dfa_toNode->id+1;
      clow = edge->dfa_upper+1;
      edgcnt += 1;
    }
    if (clow < SCN_CHAR_SET_SIZE)
    {
      EdgeC[edgcnt] = clow; EdgeS[edgcnt] = 0; edgcnt += 1;
    }
    for (i = 0; i < (edgcnt-StaEdg[node->id])/2; i ++)
    { short t;
      t = EdgeC[StaEdg[node->id]+i];
          EdgeC[StaEdg[node->id]+i] = EdgeC[edgcnt-1-i];
                                      EdgeC[edgcnt-1-i] = (byte)t;
      t = EdgeS[StaEdg[node->id]+i];
          EdgeS[StaEdg[node->id]+i] = EdgeS[edgcnt-1-i];
                                      EdgeS[edgcnt-1-i] = t;
    }
  }
  StaEdg[states] = edgcnt;
  /* make scanner structure */
  res = New(scn_t);
  res->Name   = Scn_name;
  res->States = states;
  res->Tokens = topEntry;
  res->StaEdg = StaEdg;
  res->StaFin = StaFin;
  res->EdgeC  = EdgeC;
  res->EdgeS  = EdgeS;
  res->TokId  = (string*)NewMem((topEntry)   * sizeof(string));
  res->Flags  = (byte  *)NewMem((topEntry)   * sizeof(byte));
  for (i = 0; i < topEntry; i++)
    res->TokId[i] = StrCopy(Table[i].name);
  for (i = 0; i < topEntry; i++)
    res->Flags[i] = Table[i].usage;
  /* free temporaries */
  RegSet_Free(scanner);
  /* done */
  return Scn_ADT(res);
}

/* -- graph printer -- */

static void pSource(faNode x)
{
  fprintf(STDOUT,"%d(%s)", x->Belongs->id,Table[x->Belongs->id].name);
}

static void pChar(byte c)
{
  switch (c)
  {
    case ' '  : fprintf(STDOUT,"\\_"); break;
    case '\f' : fprintf(STDOUT,"\\p"); break;
    case '\n' : fprintf(STDOUT,"\\n"); break;
    case '\r' : fprintf(STDOUT,"\\r"); break;
    case '\\' :
    case '\'' :
    case '\"' :
    case '`'  : fprintf(STDOUT,"\\%c",c); break;
    default : if (32 < c && c < 127)
                fprintf(STDOUT,"%c ",c);
              else
                fprintf(STDOUT,"%02x",c);
  }
}

static void pEdge(dfaEdge x)
{
  fprintf(STDOUT,"(");
  if (x->dfa_lower != x->dfa_upper)
  { pChar((byte)x->dfa_lower); fprintf(STDOUT," - "); }
  else
    fprintf(STDOUT,"     ");
  pChar((byte)x->dfa_upper);
  fprintf(STDOUT," : %2d)", x->dfa_toNode->id );
}

static void pNfaEdge(nfaEdge x)
{
  fprintf(STDOUT,"({\'%c\'..\'%c\'}, ", x->nfa_lower, x->nfa_upper);
  pList(x->nfa_toNodes, -1, (void (*)(Abs_T x)) pSource);
  fprintf(STDOUT,")");
}

static void pNfaEdges(List(nfaEdge) x)
{
  pList(x,-1, (void (*)(Abs_T x)) pNfaEdge);
}

static void pNode(faNode x)
{
  fprintf(STDOUT,"%2d %2d ", x->id, x->term);
  fprintf(STDOUT,"(%2d %2d) ", x->flag1, x->flag2);
  pList(x->Source,-1, (void (*)(Abs_T x)) pSource);
  fprintf(STDOUT,"\n      ");
  pList(x->Edges,-1, (void (*)(Abs_T x)) pEdge);
}

void RegSet_Print(RegSet_T x)
{
  fprintf(STDOUT,"RegSet_T %d =\n",RegSet_IMP(x)->id);
  pList(RegSet_IMP(x)->Nodes,0, (void (*)(Abs_T x)) pNode);
  fprintf(STDOUT,"\n");
}

Scn_T Scn_dfnEnd(bool diagnose)
/* definition complete, create scanner */
{ int i; List(RegSet_Ts) graphs = NULL;
  RegSet_T scanner; Scn_T scn;
  bug0(Scn_defining,"no definition");
  for (i = 1; i < topEntry; i++)
  {
    graphs = cons(Table[i].value,graphs);
    fst(regset_t,graphs)->id = i;
  }
  ConflictFlag = False;
  scanner = RegSet_Merge(graphs, NULL,NULL);
  terminalScanner(scanner);
  bug0(!ConflictFlag,"aborted due to conflicts");
  if (diagnose) RegSet_Print(scanner);
  scn = ScnTable(scanner);
  for (i = 1; i < topEntry; i++)
  { FreeMem(Table[i].name); RegSet_Free(Table[i].value); }
  Scn_defining = False;
  return scn;
}
