/* ------------------------------------------------------------------------ */
/*                                                                          */
/* [pim_test.c]                  Parser Test                                */
/*                                                                          */
/* Copyright (c) 1993 by D\olle, Manns                                      */
/* ------------------------------------------------------------------------ */

#include "stdosx.h"
#include "scn_base.h"
#include "scn_pre.h"
#include "prs.h"
#include "ptm.h"
#include "ptm_gen.h"
#include "pathes.h"
#include "symbols.h"
#include "hmap.h"
#include "gstream.h"

#include "pim__cim.h"

#define start    CTX_EVAL(string,"Start")
#define language CTX_EVAL(string,"Language")
#define filename CTX_EVAL(string,"FileName")
#define prepar   CTX_EVAL(string,"prepar")
#define premac   CTX_EVAL(string,"premac")
#define cset     CTX_EVAL(string,"charset")
#define wprint   CTX_EVAL(bool,"wprint")
#define tree     CTX_EVAL(bool,"tree")
#define xaron    CTX_EVAL(bool,"xaron")
#define early    CTX_EVAL(bool,"early")

int main(int argc, string argv[])
{ PLR_Tab PTab; Scn_T Scn; Scn_Stream cStream; 
  PT_Cfg PCfg; PT_Term PTree;
  MAP(long,symbol) pPreParMap = (MAP(long,symbol))NULL;
  CTX_init_pim_test(argc,argv);
  initSymbols();
  MAP_init();
  /* init pre-processing */
  if( strlen(prepar) > 0 )
  { c_string szS, szT; PIT pItr; 
    pPreParMap = MAP_newPrimMap();
    pItr = PIT_make(prepar); 
    PIT_sep(pItr,",");
    while( ( szS = PIT_read(pItr) ) != NULL )
    {
      szT = strstr(szS,"~");
      if( szT != (c_string)NULL )
      {
        *szT = '\0';
        MAP_dfndom(pPreParMap,atol(szS),stringToSymbol(szT+1));
        *szT = '~';
      }
    }
    PIT_drop(pItr);
    if( ! MAP_defined(pPreParMap,SPP_INC_CHARSET) && strlen(cset) > 0 )
      MAP_dfndom(pPreParMap,SPP_INC_CHARSET,stringToSymbol(cset));
    SPP_init(pPreParMap);
    /* add pre-defined macros */
    if( strlen(premac) > 0 )
    {
      pItr = PIT_make(prepar); 
      PIT_sep(pItr,",");
      while( ( szS = PIT_read(pItr) ) != NULL )
      {
        SPP_addCtxMacro(stringToSymbol(szS));
      }
      PIT_drop(pItr);
    }
  }
  /* get the scan stream */
  Scn_get(language,&Scn);
  if( strlen(cset) == 0 )
    cStream = Stream_file(Scn,"",filename,"");
  else
  { FILE*          pFP = OpnFile(filename,"rb");
    GStream_T      pGS = GS_stream_file_new(pFP, cset, True); 
    Scn_Stream_Itr pChrItr;
    assert1(pGS != (GStream_T)NULL, "creation of stream '%s' failed",filename);
    pChrItr = Stream_Itr_new
              (
                GS_stream_get_wcrc, GS_fun_destruct(pGS), GS_ucs4_to_utf8, 
                GS_fun_seek(pGS), pGS, filename
              );
    cStream = Stream_bgn(Scn,pChrItr);
    Stream_Itr_free(pChrItr);
  }
  if( strlen(prepar) > 0 )
    Stream_premac_set(cStream, SPP_premac);
  /* get the parser */
  PTab    = PLR_getTab(language);
  PCfg    = PT_init(PTab,cStream);
  if( ! xaron ) 
  {
    if( ! early )
      PTree = PT_PARSE(PCfg,start);
    else
    {
      PTree = PT_INTERPRETE(PCfg,start);
      while( PTree != (PT_Term)NULL )
      {
        if( tree ) 
        {
          if( strlen(cset) == 0 ) PT_print(PTree);
          else 
          if( wprint ) PT_print_utf8(PTree);
          else         PT_print_rawutf8(PTree);
        }
        PT_delT(PTree);
        PTree = PT_INTERPRETE(PCfg,start);
      }
    }
  }
  else          
  {
    if( ! early )
      PTree = XPT_PARSE(PCfg,start);
    else
    {
      PTree = XPT_INTERPRETE(PCfg,start);
      while( PTree != (PT_Term)NULL )
      {
        if( tree ) 
        {
          if( strlen(cset) == 0 ) PT_print(PTree);
          else 
          if( wprint ) PT_print_utf8(PTree);
          else         PT_print_rawutf8(PTree);
        }
        PT_delT(PTree);
        PTree = PT_INTERPRETE(PCfg,start);
      }
    }
  }
  PT_quit(PCfg);
  /* print derivation tree */
  if( tree && PTree != (PT_Term)NULL ) 
  {
    if( strlen(cset) == 0 ) PT_print(PTree);
    else 
    if( wprint ) PT_print_utf8(PTree);
    else         PT_print_rawutf8(PTree);
  }
  /* release scan stream and parser */
  Stream_close(cStream);
  Stream_free(cStream);
  Scn_free(Scn);
  PLR_delTab(PTab);
  if( PTree != (PT_Term)NULL ) PT_delT(PTree);
  /* quit pre-processing */
  if( strlen(prepar) > 0 )
  {
    MAP_freeMap(pPreParMap);
    SPP_quit();
  }
  MAP_quit();
  freeSymbols();
  CTX_quit();
  BUG_CORE;
  return(0);
}
