# [ctoh.cmd] Kommandobeschreibung fuer [ctoh.c]

Program ctoh -- Uebersetzer fuer Programmkontexte

doc
  Zusammenfassung

    "ctoh" erzeugt aus der [*.c]-Quelle eines Moduls dessen [*.h]-File.

  Verwendung

    Diese Programm wird vor [cdps] auf alle C-Modul-Quellen angewendet.

  Methode

    Die Quellen der C-Module werden dahingehend gekennzeichnet, da\s einzelne,
    aufeinanderfolgende Gruppen von Zeilen dem Interface zugeh\orig sein sollen.
    Diese Zeilengruppen werden im folgenden "Block" genannt.

    Leider ist die Konvention etwas komplizierter als ich gehofft habe.
    Die Grundidee ist, da\s ein Block mit einem Wort (dem Typ-Namen einer
    Funktionsdefinition, n\amlich) beginnt, und mit einer Leerzeile oder
    einer \offnenden, geschweiften Klammer ('{') endet.
    Letztere wird, der Einfachheit halber, in ein ";" \ubersetzt.

    Von dieser Grundregel gibt es jetzt eine Anzahl an Ausnahmen.

    Zeilen, die zwar mit einem Wort anfangen, da\s jedoch erkennbar
    nicht in das Interface geh\ort, starten keinen Block. Die sind
    'static', 'ConcreteType', 'struct', 'union' und 'typedef'.

    Im Hinblick auf C++-Quellen starten folgende Zeilen keinen Block:
    - Zeilen, die mit 'inline', 'class', 'template' beginnen
    - Zeilen, die der Beginn der Definition einer Elementfunktion sind

    Da die bisherige Regel etwas zuviel ausschlie\st, gibt es von der
    bisherigen Regel wiederum Ausnahmen.

    Der erste Block wird grunds\atztlich kopiert. Dies ist der Dateikopf,
    innerhalb dessen ggfls. die Extention des Dateinames von '.c' in '.h'
    umgepatcht wird.

    Eine Zeile mit einer '#include'-Anweisungen beginnt, startet ebenfalls
    einen Block. Diese Regel wird allerdings nur einmal angewendet. Die
    '#include'-Anweisungen, die im [*.h]-File auftreten sollen, m\ussern
    also in der Quelle zuerst kommen und durch eine Leerzeilen von dern
    anderen getrennt werden.

    Kommentarzeilen, die mit '/*I-' oder '/*I=' beginnen, starten einen Block.
    Das 'I' am Anfang wird dabei in ein Leerzeichen umgepatcht.

    Um auch alle sonstigen Zeilen \ubernehmen zu k\onnen, kann ein Block mit
    dem Kommentar '/*INTERFACE*/' gestartet werden. Diese Zeile wird nicht
    mitkopiert.

    Um zu erm\oglichen, da\s bestimmte Texte "nur" im Interface auftreten,
    ist abweichend die Form folgende Form zul\assig. Der Makroname
    'INTERFACE' darf daher nicht definiert werden. Diese Konvention
    erlaubt innerhalb des eingeschlossenen Textes auch Kommentare
    anzugeben. Die erste und letzte Zeile wird nicht mitkopiert.

  | #ifdef INTERFACE
  | ...
  | #endif

    Ein Dokumentationsblock ( s.u. ), wobei der Makroname wiederum nicht
    definiert werden darf, wird bei der Interface-Generierung \ubersprungen.

  | #ifdef DOCUMENT
  | ...
  | #endif

    Die restliche Konvention betrifft die Layoutverbesserung des resultierendern
    [*.h]-Files. Hierbei wird versucht, das Layout von Funktionsdefinitionern
    zu optimieren. Bl\ocke, deren 1. Zeile mit einem Buchstaben beginnt und
    mit einer ')' endet und deren deren zweite Zeile mit '/*' oder mit '{'
    beginnt, werden als Funktionsdefinitionen betrachtet, die in einzeilig
    zu layouten sind.

    Die Kommentarzeile '/*INTERFACE_GROUP*/' erzeugt eine Leerzeile.

    Um [make] nicht den Eindruck zu vermitteln, da\s Interface h\atte sich
    ge\andert, wird zun\achst nur eine tempor\are Datei erzeugt, die mit
    dem ggfls. schon existierenden [*.h]-File verglichen wird. Diese wird
    nur im Falle eines tats\achlichen Unterschieds \uberschrieben.

    Der Text wird durch die \ubliche Bedingung eingeschlossen.

  | <Kopf>
  | #ifndef <FILE>_INCL
  | #define <FILE>_INCL
  | <Text>
  | #endif
cod

Arguments

  CPATH    : path         -- [DLL-BaseName~]Quellverzeichnis ; ... [*.c]
  HPATH    : path         -- Zielverzeichnis                       [*.h]
  PRJ      : path         -- Verzeichnis fuer                      [ctoh.cth]

Options

  force    : flag   = false -- Erzwingt die Erzeugung eines neues [*.h]-Files
  ct       : flag   = false -- Konvertierung der Typen string,bool,byte,word
  exp      : flag   = false -- Generierung von MS-Exportanweisungen / Modul 
  api      : string = ""    -- DLL-BasisName fr DLL-Ex/Import

doc
  Option '-force'

    Normalerweise \uberpr\uft [ctoh] zun\achst, ob das [*.c]-File j\unger
    ist, als das [*.h]-File. Nur in diesem Fall (oder falls die Option
    '-force' angegeben wurde, wird dann die Quelle tats\achlich verarbeitet.
    Quellen, die eine Zeile enthalten, die mit 'main' beginnt, gelten als
    Programm und f\uhren ebenfalls nicht auf die Generierung eines [*.h]-Files.

  Option '-api'

    Die Option identifiziert die DLL.
cod

Environment
