/* ------------------------------------------------------------------------- */
/*                                                                           */
/* [styx.c]                 STYX Compiler Module                             */
/*                   - realizes an abstract syntax tree -                    */
/*                                                                           */
/* Copyright (c) 1993 by D\olle, Manns.                                      */
/* ------------------------------------------------------------------------- */

#include "stdosx.h"

#include "pathes.h"
#include "literal.h"
#include "prs.h"
#include "prs_gen.h"
#include "cfg_abs.h"
#include "ptm.h"
#include "ptm_gen.h"
#include "gls.h"
#include "symbols.h"
#include "hmap.h"
#include "hset.h"
#include "scn_base.h"
#include "scn_gen.h"
#include "scn_pre.h"
#include "sink.h"

#include "styx_cim.h"
#include "styx_int.h"
#include "styx_lim.h"
#include "styx_pim.h"

/* Type for global definitions */
static MAPTY                 gloty = (MAPTY)NULL; 
/* Global Definitions          */
static MAP(PT_Term, PT_Term) glo   = (MAP(PT_Term, PT_Term))NULL;   
/* used tokens                 */
static HS_Set                tkset = (HS_Set)NULL; 

/* semantic check */

static void Pass1(styxSource s)
/* Aufsammeln der Definitionen und Scoping Rules */
{ bool startflg = False;
  MAP(PT_Term,NULL) pro = MAP_newMap( gloty ); /* Productions */
  PT_Itr it; styxOptCfg optcfg; GLS_Lst(styxDfn) dfns;
  styxSource_root(s,_,_,&optcfg);
  PT_FORALL(it,s)
  { PT_Term t = PT_termIT(it);
    if (PT_stateIT(it) == PT_POST)
    { styxQlxDfn qd; styxDfn pd; styxPrd p; GLS_Tok id;
      /* Scope(Source): alle Bezeichner in QLX-Definitionen */
      if (styx_QlxDfn(t,&qd) )
      { styxQlxOpt o; styxQlxCat c;
        if ( styxQlxDfn_defn(qd,&c,&o,&id,_) )
        {
          if (MAP_defined(glo,id))
            PT_error
            (
              id,"identifier '%s' allready defined",
              symbolToString(GLS_Tok_symbol(id))
            );
          else MAP_dfndom(glo,id,qd);
          /* CaseIgnore nur mit Comment und Token */
          if(   styxQlxOpt_ignca(o) && 
              ! styxQlxCat_tokC(c)  && 
              ! styxQlxCat_comC(c)   )
            PT_error(o,"CaseIgnore not allowed here");
        }
      }

      /* Scope(Source): alle Bezeichner in CFG-Definitionen */
      if (styx_Dfn(t,&pd))
      { styxCat c; styxDfnOpt o;
        if ( styxDfn_defn(pd,&c,&o,&id,_) )
        { 
          if (MAP_defined(glo,id))
            PT_error
            (
              id,"identifier '%s' allready defined",
              symbolToString(GLS_Tok_symbol(id))
            );
          else MAP_dfndom(glo,id,pd);
          if (MAP_emptyMap(pro))
            PT_error
            (
              id,"production '%s' has no rules",
              symbolToString(GLS_Tok_symbol(id))
            );
          MAP_rmvall(pro);
          startflg |= styxCat_bgnC(c);
        }
      }

      /* Scope(Definition): alle Produktionsbezeichner */
      if (styx_Prd(t,&p))
      { 
        if ( styxPrd_prod(p,_,&id,_) )
        { 
          if (MAP_defined(pro,id))
            PT_error
            (
              id,"identifier '%s' allready defined",
              symbolToString(GLS_Tok_symbol(id))
            );
          else MAP_dfndom(pro,id,NULL);
        }
      }
    }
  }
  if( ! startflg && styxOptCfg_cfg(optcfg,&dfns) && ! GLS_EMPTY(dfns) ) 
    PT_error(dfns,"missing start production");
  MAP_freeMap( pro );
}

static void Pass2(styxSource s)
/* Referenzen */
{ PT_Itr it;
  PT_FORALL(it,s)
  { PT_Term t = PT_termIT(it);
    if (PT_stateIT(it) == PT_POST)
    { styxExp exp; styxMbr m; GLS_Tok id;
      if (styx_Exp(t,&exp) )
      {
        if ( styxExp_ident(exp,&id) )
        {
          if (!MAP_defined(glo,id))
            PT_error
            (
              id,"identifier '%s' undefined",
              symbolToString(GLS_Tok_symbol(id))
            );
          else
            if (!styx_QlxDfn(MAP_apply(PT_Term,glo,id),_))
              PT_error
              (
                id,"'%s' is no QLX-identifier",
                symbolToString(GLS_Tok_symbol(id))
              );
        }
      }
      if (styx_Mbr(t,&m) )
      { 
        if ( styxMbr_ntm(m,&id) )
        { 
          if (!MAP_defined(glo,id))
            PT_error
            (
              id,"identifier '%s' undefined",
              symbolToString(GLS_Tok_symbol(id))
            );
          else
          { PT_Term def = MAP_apply(PT_Term,glo,id); 
            styxQlxDfn qd; styxQlxCat qc;
            if( ! ( styx_Dfn(def,_)                 ||
                    ( styx_QlxDfn(def,&qd)          &&
                      styxQlxDfn_defn(qd,&qc,_,_,_) &&
                      styxQlxCat_tokC(qc)            ) ) )
              PT_error
              (
                id,"'%s' is no grammar symbol",
                symbolToString(GLS_Tok_symbol(id))
              );
          }
        }
      }
    }
  }
}

/* Scanner Translation */

#define UsageLet -100
#define UsageIgn 1
#define UsageTok 2
#define UsageCom 3

static MAP(GLS_Tok,RegSet_T) qlx; /* Computed RegSets */

static RegSet_T trans_Ide(GLS_Tok id);
static RegSet_T trans_Exp(styxExp e);

static RegSet_T TRANS_RegSet_String(string s)
{ RegSet_T reg; int len; c_bool ucs4;
  wc_string tstring = LIT_styx_lit_str(&len,s,&ucs4);
  reg = RegSet_StringN(tstring,len);
  FreeMem(tstring); 
  return reg;
}

static RegSet_T TRANS_RegSet_Cset(string s)
{ RegSet_T reg; int len; c_bool ucs4;
  wc_string tstring = LIT_styx_lit_str(&len,s,&ucs4);
  reg = RegSet_CsetN(tstring,len);
  FreeMem(tstring); return reg;
}

static string TRANS_Seq(GLS_Tok seq)
{ wc_string tstring; int len; c_bool ucs4;
  string s = symbolToString(GLS_Tok_symbol(seq)), res;
  tstring = LIT_styx_lit_str(&len,s,&ucs4);
  bug0(ucs4 == C_False,"unicode characters not allowed in keywords");
  res = WCToStrCopy(tstring);
  bug0(strlen(res) == len,"'\\0' not allowed in keywords");
  FreeMem(tstring);
  return res;
}

static int qlx_usage(GLS_Tok id)
{ styxQlxCat cat;
  styxQlxDfn_defn(MAP_apply(styxQlxDfn,glo,id),&cat,_,_,_);
  if (styxQlxCat_letC(cat)) return UsageLet;
  if (styxQlxCat_tokC(cat)) return UsageTok;
  if (styxQlxCat_comC(cat)) return UsageCom;
  if (styxQlxCat_ignC(cat)) return UsageIgn;
  BUG; return 0;
}

static RegSet_T trans_Ide(GLS_Tok id)
{
  if (!MAP_defined(qlx,id))
  { RegSet_T res,old; styxExp exp;
    MAP_dfndom(qlx,id,NULL); /* Flag to protect against endless recursion */
    styxQlxDfn_defn(MAP_apply(styxQlxDfn,glo,id),_,_,_,&exp);
    res = trans_Exp(exp);
    old = MAP_apply(RegSet_T,qlx,id);
    if (old != (RegSet_T)NULL) RegSet_Free(old);
    MAP_upddom(qlx,id,res);
  }
  else
  if (MAP_apply(RegSet_T,qlx,id) == (RegSet_T)NULL)
  {
    PT_error(id,"recursive definition");
    MAP_upddom(qlx,id,RegSet_Empty());
  }
  return RegSet_Copy(MAP_apply(RegSet_T,qlx,id));
}

static RegSet_T trans_Exp(styxExp e)
{ styxExp exp1; styxExp exp2; GLS_Tok tok;
  if( styxExp_union(e,&exp1,&exp2) )
  { RegSet_T a = trans_Exp(exp1);
    RegSet_T b = trans_Exp(exp2);
    RegSet_T res = RegSet_Union(a,b);
    RegSet_Free(a); RegSet_Free(b);
    return res;
  }
  if( styxExp_diff(e,&exp1,&exp2) )
  { RegSet_T a = trans_Exp(exp1);
    RegSet_T b = trans_Exp(exp2);
    RegSet_T res = RegSet_Difference(a,b);
    RegSet_Free(a); RegSet_Free(b);
    return res;
  }
  if( styxExp_conc(e,&exp1,&exp2) )
  { RegSet_T a = trans_Exp(exp1);
    RegSet_T b = trans_Exp(exp2);
    RegSet_T res;
    res = RegSet_Concat(a,b);
    RegSet_Free(a); RegSet_Free(b);
    return res;
  }
  if( styxExp_opt(e,&exp1) )
  { RegSet_T a = trans_Exp(exp1);
    RegSet_T res = RegSet_Option(a);
    RegSet_Free(a);
    return res;
  }
  if( styxExp_star(e,&exp1) )
  { RegSet_T a = trans_Exp(exp1);
    RegSet_T res = RegSet_Star(a);
    RegSet_Free(a);
    return res;
  }
  if( styxExp_plus(e,&exp1) )
  { RegSet_T a = trans_Exp(exp1);
    RegSet_T res = RegSet_Plus(a);
    RegSet_Free(a);
    return res;
  }
  if( styxExp_range(e,&exp1,&exp2) )
  { RegSet_T a = trans_Exp(exp1);
    RegSet_T b = trans_Exp(exp2);
    RegSet_T res;
    short al, bu;
    if (!RegSet_isChar(a)) PT_error(exp1,"single character");
    if (!RegSet_isChar(b)) PT_error(exp2,"single character");
    al = RegSet_charVal(a);
    bu = RegSet_charVal(b);
    if (al > bu)
    { short x = al; al = bu; bu = x; PT_error(e,"range must be ordered"); }
    res = RegSet_Range(al,bu);
    RegSet_Free(a); RegSet_Free(b);
    return res;
  }
  if( styxExp_set(e,&tok) )
    return TRANS_RegSet_Cset(symbolToString(GLS_Tok_symbol(tok)));
  if( styxExp_sequ(e,&tok) )
    return TRANS_RegSet_String(symbolToString(GLS_Tok_symbol(tok)));
  if( styxExp_ident(e,&tok) )
    return trans_Ide(tok);
  BUG; return (RegSet_T)NULL;
}

static Scn_T trans_QlxDfns(styxSource s, bool diagnose)
{ GLS_Tok id; Scn_T scn; PT_Itr it; MAPIT itr;
  qlx = MAP_newMap(gloty);
  styxSource_root(s,&id,_,_);
  Scn_dfnBegin(symbolToString(GLS_Tok_symbol(id)));
  PT_FORALL(it,s) if (PT_stateIT(it) == PT_PRAE)
  { PT_Term t = PT_termIT(it); styxQlxDfn d;
    if( styx_QlxDfn(t,&d) )
    { styxQlxOpt o;
      if( styxQlxDfn_defn(d,_,&o,&id,_) )
      {
        if( qlx_usage(id) == UsageTok ||
            qlx_usage(id) == UsageIgn ||
            qlx_usage(id) == UsageCom  )
        { RegSet_T a = trans_Ide(id);
          Scn_dfnToken
          (
            symbolToString(GLS_Tok_symbol(id)),
            (byte)( qlx_usage(id) == UsageIgn
                  ? SCN_FLG_IgnoreToken
                  : 0 | ( styxQlxOpt_ignca(o) ? SCN_FLG_IgnoreCase : 0)
                  ),
            a
          );
          RegSet_Free(a);
        }
      }
    }
  }
  scn = Scn_dfnEnd(diagnose);
  MAP_forItr(id,itr,qlx) RegSet_Free(MAP_apply(RegSet_T,qlx,id));
  MAP_freeMap(qlx);
  return scn;
}

/* checking keywords */

static void Pass3(Scn_T scn, styxSource s)
/* keywords */
{ PT_Itr it;
  PT_FORALL(it,s) if (PT_stateIT(it) == PT_POST) /* alle reservierten Worte */
  { PT_Term t = PT_termIT(it); styxMbr m; GLS_Tok seq;
    if( styx_Mbr(t,&m) && styxMbr_tkm(m,&seq) )
    { string s = TRANS_Seq(seq);
      /*FIXME: comments are not allowed, too*/
      if (!Scn_check_Token(scn,s)) PT_error(seq,"malformed token");
      FreeMem(s);
    }
  }
}

/* Parser Translation */

static string normalKeyword(Scn_T scn, GLS_Tok seq)
/* normalizes keywords (IgnoreCase) */
{ string s = TRANS_Seq(seq); short tid = Scn_check_Token(scn,s);
  bug1(tid,"token '%s' does not belong to grammar\n",s);
  if( Scn_tokFlags(scn,tid) & SCN_FLG_IgnoreCase ) StrToLower(s);
  return s;
}

static PLR_Cfg trans_Dfns(styxSource s, Scn_T scn)
{ MAPTY keyty; MAP(symbol,NULL) key; /* reserved words */
  PLR_Cfg Cfg; MAPIT itr; PT_Itr it; 
  GLS_Lst(styxDfn) dfns; styxOptCfg optcfg;
  styxDfn pd; styxMbr m; GLS_Tok id; GLS_Tok seq;
  styxSource_root(s,&id,_,&optcfg);
  styxOptCfg_cfg(optcfg,&dfns);
  Cfg   = PLR_createCfg(symbolToString(GLS_Tok_symbol(id)));
  keyty = MAP_newTyp(primCopy,primFree,primEqual,primHash,primCopy,primFree);
  key   = MAP_newMap(keyty);
  PT_FORALL(it,s) if( PT_stateIT(it) == PT_PRAE )
  { PT_Term t = PT_termIT(it); styxQlxDfn qd; 
    if( styx_QlxDfn(t,&qd) ) /* Token */
    { 
      if( styxQlxDfn_defn(qd,_,_,&id,_)                              &&
          ( qlx_usage(id) == UsageTok || qlx_usage(id) == UsageCom )  )
          PLR_addTK(Cfg,symbolToString(GLS_Tok_symbol(id)),PLR_TYP_TOK);
    }
    if( styx_Dfn(t,&pd) ) /* Nonterminal */
    { styxDfnOpt o;
      if( styxDfn_defn(pd,_,&o,&id,_) )
        PLR_addNT(Cfg,symbolToString(GLS_Tok_symbol(id)),styxDfnOpt_errnt(o));
    }
    if( styx_Mbr(t,&m) ) /* Keyword */
    { 
      if( styxMbr_tkm(m,&seq) )
      { string s       = normalKeyword(scn,seq); 
        symbol keyword = stringToSymbol(s);
        FreeMem(s);
        if( ! MAP_defined(key,keyword) )
        {
          MAP_dfndom(key,keyword,NULL);
          PLR_addTK(Cfg,symbolToString(keyword),PLR_TYP_KEY);
        }
      }
      else
      if( styxMbr_else(m) )
      {
        PLR_addTK(Cfg,PLR_TOK_Other,PLR_TYP_TOK);
      }     
    }
  }
  MAP_freeMap( key );
  MAP_freeTyp( keyty );

  PLR_endSD(Cfg); /* Sorting Token & Nonterminals */

  MAP_forItr(id,itr,glo)
    if( styx_QlxDfn(MAP_apply(PT_Term,glo,id),_) && 
        qlx_usage(id) == UsageCom                 ) /* Special Token */
      PLR_addST(Cfg,symbolToString(GLS_Tok_symbol(id)));
  PT_FORALL(it,dfns) if( PT_stateIT(it) == PT_PRAE )
  { PT_Term t = PT_termIT(it); GLS_Tok nt; styxPrd p;
    if( styx_Dfn(t,&pd) ) /* Startsymbol */
    { styxCat c;
      if( styxDfn_defn(pd,&c,_,&nt,_) && styxCat_bgnC(c) )
        PLR_addSN(Cfg,symbolToString(GLS_Tok_symbol(nt)));
    }

    if( styx_Prd(t,&p) ) /* Production */
    { styxLay lay; GLS_Tok pid;
      if( styxPrd_prod(p,&lay,&pid,_) )
      {
        if( styxLay_dft(lay) )
          PLR_addPR
          (
            Cfg,symbolToString(GLS_Tok_symbol(pid)),
            PLR_PRD_DFT,symbolToString(GLS_Tok_symbol(nt))
          );
        else
        if( styxLay_rec(lay) )
          PLR_addPR
          (
            Cfg,symbolToString(GLS_Tok_symbol(pid)),
            PLR_PRD_REC,symbolToString(GLS_Tok_symbol(nt))
          );
        else
        if( styxLay_grp(lay) )
          PLR_addPR
          (
            Cfg,symbolToString(GLS_Tok_symbol(pid)),
            PLR_PRD_GRP,symbolToString(GLS_Tok_symbol(nt))
          );
      }
    }
    if( styx_Mbr(t,&m) ) /* Production Symbol */
    { int r = PT_row(m), c = PT_col(m);
      if( styxMbr_tkm(m,&seq) )
      { string s = normalKeyword(scn,seq);
        PLR_addPT(Cfg,s,r,c);
        FreeMem(s);
      }
      else
      if( styxMbr_ntm(m,&id) )
      { PT_Term t = MAP_apply(PT_Term,glo,id);
        r = PT_row(id);
        c = PT_col(id);
        if( styx_QlxDfn(t,_) )
        {
          PLR_addPT(Cfg,symbolToString(GLS_Tok_symbol(id)),r,c);
          HS_SET_ELM(GLS_Tok_symbol(id),tkset);
        }
        if( styx_Dfn(t,_) )
          PLR_addPN(Cfg,symbolToString(GLS_Tok_symbol(id)),r,c);
      }
      else
      if( styxMbr_else(m) )
      {
        PLR_addPT(Cfg,PLR_TOK_Other,r,c);
        HS_SET_ELM(stringToSymbol(PLR_TOK_Other),tkset);
      }
    }
  }
  return Cfg;
}

/* c interface generation */

#define INTEXT "int" /* FileNames will be LANGUAGE_INTEXT.(c|h) */
#define ANTEXT "ant" /* FileNames will be LANGUAGE_ANTEXT.(c|h) */

#define LineLen 78
#define SNL FNL(SysOut)

static FILE *SysOut;
static PLR_Tab PTab;

/* abstract syntax data structure ( see comments in cfg_abs.c ) */
static int*    aPol;
static HS_Set* aNtm;
static HS_Set* aPrd;

static int  maxlen = 0;
static char tkclsnam[6]; /* token class name */
static char lprefix[5];  /* language prefix  */

static bool genabs; /* True: generate for row representation */

static void nChar(FILE* f, int n, char c)
/* n * c */
{ int i;
  for( i = 0; i < n; i++) fprintf(f,"%c",c);
}

static void fileComment(FILE *f, string filename, string s)
/* layout a file comment */
{
  fprintf(f,"/* "); nChar(f,LineLen-6,'-'); fprintf(f," */"); FNL(f);
  fprintf(f,"/* "); nChar(f,LineLen-6,' '); fprintf(f," */"); FNL(f);
  fprintf(f,"/* [%s]",filename);
  nChar(f,(LineLen-6-strlen(s)+0)/2-2-strlen(filename),' ');
  fprintf(f,"%s",s);
  nChar(f,(LineLen-6-strlen(s)+1)/2,' ');
  fprintf(f," */"); FNL(f);
  fprintf(f,"/* "); nChar(f,LineLen-6,' '); fprintf(f," */"); FNL(f);
  fprintf(f,"/* "); nChar(f,LineLen-6,'-'); fprintf(f," */"); FNL(f);
  FNL(f);
}

static void sectionComment(string s)
/* layout a section comment */
{
  SNL;
  fprintf(SysOut,"/*I");
  nChar(SysOut,(LineLen-6-strlen(s)-2+0)/2,'-');
  fprintf(SysOut," %s ",s);
  nChar(SysOut,(LineLen-6-strlen(s)-2+1)/2,'-');
  fprintf(SysOut," */"); SNL;
}

/* ------- Abstract Grammar Production ------ */

static void genNtmMbrComment(FILE* f, int clsidx)
/* generates nonterminal member comment */
{
  if( aPol[clsidx] == -1 )
    fprintf(f,"%s",PLR_symName(PTab,PLR_ntClass(PTab,clsidx)));
  else
  {
    if( aPol[clsidx] >= 0 )
      genNtmMbrComment(f,aPol[clsidx]); 
    else
    {
      if( ABS(aPol[clsidx]+2) == PLR_tokenCnt(PTab) )
        fprintf(f,"GLS_Tok");
      else
      {
        fprintf(f,"%s",PLR_symName(PTab,ABS(aPol[clsidx]+2))); 
      }
    }
    fprintf(f,"*");
  }
}

static void genProdComment(FILE* f, int i, int indent)
/* generates production comment */
{ int cnt = PLR_prodSymCnt(PTab,i), j; bool flg = True;
  if( indent ) fprintf(f,"  %*s",indent,"");
  fprintf(f,"%s",PLR_prodName(PTab,i));
  for( j=0; j < cnt; ++j )
  { int sym = PLR_prodSymbol(PTab,i,j), typ = PLR_symType(PTab,sym);
    if( CFG_TOK_NTM(typ) )
    {
      if( flg == True ) { flg = False; fprintf(f,"("); }
      else fprintf(f,", ");
      if( CFG_TOK(typ) ) fprintf(f,"%s",PLR_symName(PTab,sym));
      if( CFG_NTM(typ) )
        genNtmMbrComment(f,PLR_ntClassId(PTab,sym-PLR_tokenCnt(PTab)));
    }
  }
  fprintf(f,"%s", ( flg ) ? "" : ")");
}

static void genTokenCom(FILE* f)
/* generates comment for tokens */
{ int cnt = PLR_tokenCnt(PTab), i; bool flg = True;
  fprintf(f,"TOKENS"); FNL(f);
  FNL(f);
  fprintf(f,"  ");
  for( i=0; i < cnt; ++i ) 
  {
    if( CFG_TOK(PLR_symType(PTab,i))                          &&
        HS_MBR_ELM(stringToSymbol(PLR_symName(PTab,i)),tkset)  )
    {
      if( flg == True ) flg = False;
      else              fprintf(f,", ");
      fprintf(f,"%s",PLR_symName(PTab,i));
    }
  }
  FNL(f);
}

static void genTypesCom(void)
/* generates comment for types */
{ int scnt = PLR_startCnt(PTab), ncnt = PLR_nontermCnt(PTab), i;
  FILE* f = OpenPath("GENSTYX",PLR_language(PTab),".abs","wt");
  string filename = Str_printf("%s.abs",PLR_language(PTab));
  maxlen = MAX(maxlen,strlen(PLR_language(PTab)));
  for (i = 0; i < PLR_nontermCnt(PTab); i++) if (PLR_ntClassId(PTab,i) == i)
      maxlen = MAX
               (
                 maxlen,
                   strlen(PLR_symName(PTab,PLR_ntClass(PTab,i)))
                 + strlen(lprefix)
               );
  fileComment(f,filename,"Abstract Grammar");
  fprintf(f,"LANGUAGE %s",PLR_language(PTab)); FNL(f);
  FNL(f);
  genTokenCom(f);
  FNL(f);
  fprintf(f,"TYPES"); FNL(f);
  FNL(f);
  fprintf(f,"  %-*s = ",maxlen,PLR_language(PTab));
  /* 'start' productions */
  for (i = 0; i < scnt; i++) 
  {
    genProdComment(f,PLR_prodCnt(PTab)-scnt+i, ( ! i ) ? 0 : maxlen + 3);
    if( i < scnt - 1 ) fprintf(f,";"); FNL(f);
  }
  /* normal productions */
  for (i = 0; i < ncnt; i++) 
    if (PLR_ntClassId(PTab,i) == i && aPol[i] == -1 && ! HS_EMPTY_SET(aPrd[i]))
    { HS_Itr itr; HS_Elm elm; int cnt = 0, indent = 0;
      FNL(f);
      fprintf
      (
        f,"  %-*s = ",maxlen,PLR_symName(PTab,PLR_ntClass(PTab,i))
      );
      /* Here is some stuff to report class equivalence.  */
      /* We dont use it, but must eventually activate it. */
      /*
      { int j; int ntcls = PLR_ntClass(PTab,i);
        for (j = 0; j < ncnt; j++) 
        { int nt = PLR_NTI_TO_NTS(PTab,j), equiv = PLR_NTS_TO_NTC(PTab,nt);
          if( ntcls == equiv ) fprintf(f," %s",PLR_symName(PTab,nt));
        }
      }
      */
      HS_FORALL(elm,itr,aPrd[i]) 
      {
        genProdComment(f,(int)elm,indent);
        ++cnt; if( cnt < HS_CARD(aPrd[i]) ) fprintf(f,";"); FNL(f);
        if( ! indent ) indent = maxlen + 3;
      }
    }
  FNL(f);
/*
  fprintf(f,"  X*%*s = nil; cons(X,X*)",MAX(0,maxlen-2),""); FNL(f);
  FNL(f); 
*/
  fclose(f);
  FreeMem(filename);
}

/* ------------------ end of Abstract Grammar ---------------- */

static void genNtmArgTyp(int clsidx)
/* generates nonterminal argument type name */
{
  if( aPol[clsidx] == -1 )
    fprintf(SysOut,"%s%s",lprefix,PLR_symName(PTab,PLR_ntClass(PTab,clsidx)));
  else
  {
    fprintf(SysOut,"GLS_Lst(");
    if( aPol[clsidx] >= 0 )
      genNtmArgTyp(aPol[clsidx]);
    else fprintf(SysOut,"GLS_Tok");
    fprintf(SysOut,")");
  }
}

static void genProd(int i)
/* generates for Production i */
{ int nt     = PLR_prodNonTerm(PTab,i), 
      clsidx = PLR_ntClassId(PTab,nt-PLR_tokenCnt(PTab)), 
      ntcls  = PLR_ntClass(PTab,clsidx), 
      num    = 0, cnt, len, j;
  HS_Itr itr; HS_Elm elm; bool flg = True;
  len =   7 + strlen(lprefix) + strlen(PLR_symName(PTab,ntcls))
        + 1 + strlen(PLR_prodName(PTab,i)) + 1;
  SNL;
  fprintf
  (
    SysOut,"c_bool %s%s_%s(%s%s x", 
    lprefix, PLR_symName(PTab,ntcls), PLR_prodName(PTab,i), 
    lprefix, PLR_symName(PTab,ntcls)
  );
  cnt = PLR_prodSymCnt(PTab,i);
  for( j=0; j < cnt; ++j )
  { int sym = PLR_prodSymbol(PTab,i,j), typ = PLR_symType(PTab,sym); 
    if( CFG_TOK(typ) ) 
    {
      if( num > 1 && ! ( num % 2 ) )
        fprintf(SysOut,",\n%*sGLS_Tok* x%d",len,"",++num);
      else
        fprintf(SysOut,", GLS_Tok* x%d",++num);
    }
    if( CFG_NTM(typ) )
    { 
      if( num > 1 && ! ( num % 2 ) ) fprintf(SysOut,",\n%*s",len,"");
      else                           fprintf(SysOut,", ");
      genNtmArgTyp(PLR_ntClassId(PTab,sym-PLR_tokenCnt(PTab)));
      fprintf(SysOut,"* x%d",++num);
    }
  }
  fprintf(SysOut,")"); 
  SNL; fprintf(SysOut,"{"); 
  SNL; fprintf(SysOut,"  assert0( ");
  HS_FORALL(elm,itr,aNtm[clsidx])
  {
    if( flg )
    {
      flg = False;
    }
    else
    {
      SNL; fprintf(SysOut,"        ||");
    }
    if (genabs)
      fprintf(SysOut," PT_Abs_nonterm(x) == CfgSyms[%d]",(int)elm);
    else
      fprintf(SysOut," PT_THISNTM((PT_Term)x, CfgSyms[%d])",(int)elm);
  }
  fprintf(SysOut,", \"%s expected\" );",PLR_symName(PTab,ntcls));
  SNL; 
  if (genabs)
    fprintf
    (
      SysOut,"  if( PT_Abs_product(x) != PrdSyms[%d] ) return False;",i
    );
  else
    fprintf
    (
      SysOut,"  if( ! PT_THISPRD((PT_Term)x, PrdSyms[%d]) ) return False;",i
    );
  if( num )
  { bool xflg = False;
    if( !genabs )
    {
      SNL; fprintf(SysOut,"  { PT_Term y;"); 
    }
XARON_STEP:
    if( !genabs )
    {
      if( ! xflg )
      {
        SNL; fprintf(SysOut,"    if( ! PT_isXaron(x) )"); 
      }
      else
      {
        SNL; fprintf(SysOut,"    else"); 
      }
      SNL; fprintf(SysOut,"    {"); 
    }
    for( num = 0, j=0; j < cnt; ++j )
    { int sym = PLR_prodSymbol(PTab,i,j), typ = PLR_symType(PTab,sym);
      if( CFG_TOK_NTM(typ) )
      {
        SNL; 
        if (!genabs)
        {
          if( ! xflg )
            fprintf
            (
              SysOut,"      x = (%s%s)GLS_keycom_Skip(PT_%s(x));",
              lprefix, PLR_symName(PTab,ntcls),
              num ? "next" : "parts"
            );
          else
          {
            fprintf
            (
              SysOut,"      x = XGLS_keycom_Skip(%s(x));",
              num ? "rst" : "XPT_parts"
            );
            fprintf
            (
              SysOut," y = fst(%s%s,x);",lprefix,PLR_symName(PTab,ntcls)
            );
          }
          SNL; 
        }
        ++num;
        fprintf
        (
          SysOut,
          "  %sif( (%sCPtr)x%d != (%sCPtr)NULL )",
          genabs?"":"    ", genabs?"Huge":"Std", num, genabs?"Huge":"Std"
        );
        if (!genabs) { SNL; fprintf(SysOut,"       "); }
        if( CFG_TOK(typ) )
          fprintf(SysOut," *x%d = (GLS_Tok)", num);
        else
        {
          fprintf(SysOut," *x%d = (", num);
          genNtmArgTyp(PLR_ntClassId(PTab,sym-PLR_tokenCnt(PTab)));
          fprintf(SysOut,")");
        }
        if (genabs)
          fprintf(SysOut,"PT_Abs_part(x,%d);",num-1);
        else
          fprintf(SysOut,"GLS_ign_Skip(%s);",( ! xflg ) ? "x" : "y");
      }
    }
    if( !genabs )
    {
      SNL; fprintf(SysOut,"    }"); 
      xflg = ( ! xflg ) ? True : False;
      if( xflg ) goto XARON_STEP;
      SNL; fprintf(SysOut,"  }"); 
    }
  }
  SNL; fprintf(SysOut,"  return True;");
  SNL; fprintf(SysOut,"}"); 
  SNL;
}

static void genNonTerm(int i)
/* generates for nonterminal class i */
{ int nt = PLR_ntClass(PTab,i); HS_Elm elm; HS_Itr itr;
  sectionComment(PLR_symName(PTab,nt));
  HS_FORALL(elm,itr,aPrd[i]) genProd((int)elm);
}

static void genTermTyp(int i)
/* generates for terminal i */
{
  SNL; fprintf(SysOut,"c_bool %s_%s(GLS_Tok x)",tkclsnam,PLR_symName(PTab,i));
  SNL; fprintf(SysOut,"{");
  SNL; 
  if (genabs)
    fprintf(SysOut,"  return PT_Abs_token(x) == CfgSyms[%d];",i);
  else
    fprintf(SysOut,"  return( PT_THISTOK((PT_Term)x, CfgSyms[%d]) );",i);
  SNL; fprintf(SysOut,"}"); 
  SNL;
}

static void genStartTrm(void)
/* generates for startsymbol class term */
{ int nt = PLR_prodNonTerm(PTab,PLR_prodCnt(PTab)-PLR_startCnt(PTab));
  SNL; 
  fprintf
  (
    SysOut,"c_bool %s_%s(PT_%sTerm x, %s* x1)",
    lprefix, PLR_language(PTab), genabs?"Abs_":"", PLR_language(PTab)
  );
  SNL; fprintf(SysOut,"{");
  SNL;
  if (genabs)
    fprintf(SysOut,"  if( PT_Abs_nonterm(x) == CfgSyms[%d] )",nt);
  else
    fprintf(SysOut,"  if( PT_THISNTM(x, CfgSyms[%d]) )",nt);
  SNL; fprintf(SysOut,"  {");
  SNL; 
  fprintf
  (
    SysOut,"    if( (%sCPtr)x1 != (%sCPtr)NULL ) *x1 = (%s)x;", 
    genabs?"Huge":"Std",genabs?"Huge":"Std",PLR_language(PTab)
  );
  SNL; fprintf(SysOut,"    return True;");
  SNL; fprintf(SysOut,"  }");
  SNL; fprintf(SysOut,"  return False;");
  SNL; fprintf(SysOut,"}"); 
  SNL;
}

static void genNtmTrm(int i)
/* generates for nonterminal class term i */
{ int nt = PLR_ntClass(PTab,i); bool flg = True; HS_Elm elm; HS_Itr itr;
  SNL; 
  fprintf
  (
    SysOut,"c_bool %s_%s(PT_%sTerm x, %s%s* x1)", 
    lprefix, PLR_symName(PTab,nt), genabs?"Abs_":"", 
    lprefix, PLR_symName(PTab,nt)  
  );
  SNL; fprintf(SysOut,"{");
  SNL; fprintf(SysOut,"  if( ");
  HS_FORALL(elm,itr,aNtm[i])
  {
    if( flg )
    {
      flg = False;
    }
    else
    {
      SNL; fprintf(SysOut,"    ||");
    }
    if (genabs)
      fprintf(SysOut," PT_Abs_nonterm(x) == CfgSyms[%d]",(int)elm);
    else
      fprintf(SysOut," PT_THISNTM(x, CfgSyms[%d])",(int)elm);
  }
  fprintf(SysOut," )");
  SNL; fprintf(SysOut,"  {");
  SNL; 
  fprintf
  (
    SysOut,"    if( (%sCPtr)x1 != (%sCPtr)NULL ) *x1 = (%s%s)x;",
    genabs?"Huge":"Std",genabs?"Huge":"Std",
    lprefix, PLR_symName(PTab,nt)
  );
  SNL; fprintf(SysOut,"    return True;");
  SNL; fprintf(SysOut,"  }");
  SNL; fprintf(SysOut,"  return False;");
  SNL; fprintf(SysOut,"}"); 
  SNL;
}

static void genStart(int i)
/* generates for start production i */
{ int prod    = PLR_prodCnt(PTab)-PLR_startCnt(PTab)+i,
      nt      = PLR_prodNonTerm(PTab,prod),
      ntstart = PLR_startSymbol(PTab,i),
      ntclsid = PLR_ntClassId(PTab,ntstart-PLR_tokenCnt(PTab));
  SNL; 
  fprintf
  (
    SysOut,"c_bool %s_%s(%s x, ",
    PLR_language(PTab), PLR_prodName(PTab,prod), PLR_language(PTab)
  );
  genNtmArgTyp(ntclsid);
  fprintf(SysOut,"* x1)");
  /* for backward compatibility purposes ... */
  SNL; 
  fprintf
  (
    SysOut, "#define %s_Start_%d   %s_%s",
    PLR_language(PTab), i, PLR_language(PTab), PLR_prodName(PTab,prod)
  );
  /* ... for backward compatibility purposes */
  SNL; fprintf(SysOut,"{");
  SNL; 
  if (genabs)
    fprintf
    (
      SysOut,
      "  assert0( PT_Abs_nonterm(x) == CfgSyms[%d], \"%s expected\" );",
      nt, PLR_language(PTab)
    );
  else
    fprintf
    (
      SysOut,
      "  assert0( PT_THISNTM((PT_Term)x, CfgSyms[%d]), \"%s expected\" );",
      nt, PLR_language(PTab)
    );
  SNL; 
  if (genabs)
    fprintf
    (
      SysOut,"  if( PT_Abs_product(x) != PrdSyms[%d] ) return False;", prod
    );
  else
    fprintf
    (
      SysOut,"  if( ! PT_THISPRD((PT_Term)x, PrdSyms[%d]) ) return False;", prod
    );
  SNL;
  fprintf
  (
    SysOut,
    "  if( (%sCPtr)x1 != (%sCPtr)NULL )",
    genabs?"Huge":"Std",genabs?"Huge":"Std"
  );
  SNL; fprintf(SysOut,"    *x1 = (");
  genNtmArgTyp(ntclsid);
  if (genabs)
    fprintf(SysOut,")PT_Abs_part(x,0);");
  else
    fprintf
    (
      SysOut,")PT_nthSymbol((PT_Term)x, CfgSyms[%d], %d, True);",ntstart,1
    );
  SNL; fprintf(SysOut,"  return True;");
  SNL; fprintf(SysOut,"}");
  SNL;
}

static void genTypes(void)
/* generates the types */
{ int i;
  sectionComment("Types & Constants");
  SNL; 
  fprintf
  (
    SysOut,
    "Abstract%sType( %-*s );", genabs?"Huge":"",maxlen, PLR_language(PTab)
  ); 
  SNL;
  for (i = 0; i < PLR_nontermCnt(PTab); i++) 
    if (PLR_ntClassId(PTab,i) == i && aPol[i] == -1)
    {
      SNL;
      fprintf
      (
        SysOut,"Abstract%sType( %s%-*s );", genabs?"Huge":"",lprefix, 
        maxlen - strlen(lprefix), PLR_symName(PTab,PLR_ntClass(PTab,i))
      );
    }
  SNL;
  /* terminal types */
  sectionComment("Access to Tokens");
  for (i = 0; i < PLR_tokenCnt(PTab); i++) 
    if( CFG_TOK(PLR_symType(PTab,i))                          &&
        HS_MBR_ELM(stringToSymbol(PLR_symName(PTab,i)),tkset)  )
      genTermTyp(i);
  /* nonterminal types */
  sectionComment("Access to Terms");
  genStartTrm();
  for (i = 0; i < PLR_nontermCnt(PTab); i++) 
    if (PLR_ntClassId(PTab,i) == i && aPol[i] == -1) genNtmTrm(i);
  /* special type "access" */
  sectionComment(PLR_language(PTab));
  for (i = 0; i < PLR_startCnt(PTab); i++) genStart(i);
  for (i = 0; i < PLR_nontermCnt(PTab); i++) 
    if (PLR_ntClassId(PTab,i) == i && aPol[i] == -1) genNonTerm(i);
}

static void genSymbolQuit(void)
/* generates symbol objects: quit function */
{
  SNL; fprintf(SysOut,"void %s_quitSymbols()",lprefix);
  SNL; fprintf(SysOut,"{");
  SNL; fprintf(SysOut,"  if( CfgSyms != (symbol*)NULL ) FreeMem(CfgSyms);");
  SNL; fprintf(SysOut,"  if( PrdSyms != (symbol*)NULL ) FreeMem(PrdSyms);");
  SNL; fprintf(SysOut,"}");
  SNL;
}

static void genSymbolInit(void)
/* generates symbol objects: initializing function */
{ int i;
  SNL; fprintf(SysOut,"void %s_initSymbols()",lprefix);
  SNL; fprintf(SysOut,"{");
  SNL; fprintf(SysOut,"  GLS_init();");
  SNL; fprintf
       (
         SysOut,"  CfgSyms = (symbol*)NewMem(sizeof(symbol)*%d);",
         PLR_symbolCnt(PTab)
       );
  SNL; fprintf
       (
         SysOut,"  PrdSyms = (symbol*)NewMem(sizeof(symbol)*%d);",
         PLR_prodCnt(PTab)
       );
  SNL;
  for( i=0; i < PLR_symbolCnt(PTab); ++i )
  { string t = LIT_c_str_lit(PLR_symName(PTab,i));
    SNL; fprintf
         (
           SysOut,"  CfgSyms[%d] = stringToSymbol(%s);", i,t
         );
    FreeMem(t);
  }
  SNL;
  for( i=0; i < PLR_prodCnt(PTab); ++i )
  {
    SNL; fprintf
         (
           SysOut,"  PrdSyms[%d] = stringToSymbol(\"%s\");",
           i,PLR_prodName(PTab,i)
         );
  }
  SNL; fprintf(SysOut,"}"); 
  SNL;
}

static void genSymbolAccess(void)
/* generates symbol objects: initializing and quit functions */
{
  sectionComment("symbol objects - init & quit"); SNL;
  fprintf(SysOut,"static symbol* CfgSyms = (symbol*)NULL;"); SNL;
  fprintf(SysOut,"static symbol* PrdSyms = (symbol*)NULL;"); SNL;
  genSymbolInit();
  genSymbolQuit();
}

static void genInterface(void)
/* generates the interface */
{ char lid[9];
  sprintf(lprefix,"%.4s",PLR_language(PTab));
  sprintf(lid,"%s_%s",lprefix,genabs?ANTEXT:INTEXT);
  if (CTX_EVAL(bool,"makeINT") || CTX_EVAL(bool,"makeC"))
  { string filename = Str_printf("%s_%s.c",lprefix,genabs?ANTEXT:INTEXT);
    SysOut = OpenPath("GENSTYX",lid,".c","wt");
    sprintf(tkclsnam,"T%s",lprefix);
    fileComment(SysOut,filename,"Language Interface");
    fprintf(SysOut,"#include \"ptm.h\""); SNL;
    fprintf(SysOut,"#include \"gls%s.h\"",genabs?"_abs":""); SNL;
    genSymbolAccess();
    genTypes();
    fclose(SysOut);
    FreeMem(filename);
  }
}

int main(int argc, string argv[])
{ Scn_T scn; Scn_Stream cstream; PLR_Tab plr; PLR_Cfg Cfg; PT_Cfg PCfg;
  styx PTree; styxSource s; styxOptCfg optcfg; GLS_Lst(styxDfn) dfns; 
  GLS_Tok id; string language; bool verbose; bool diagnose;
  MAP(long,symbol) pPreParMap;

  CTX_init_styx(argc,argv);
  language = CTX_EVAL(string,"Language");
  verbose  = CTX_EVAL(bool,"verbose");
  diagnose = CTX_EVAL(bool,"diagnose");

  if (verbose) fprintf(STDERR,"Welcome to Styx Version '%s'\n",__DATE__);
  initSymbols();
  MAP_init();
  styx_initSymbols();
  pPreParMap = MAP_newPrimMap();
  MAP_dfndom(pPreParMap,SPP_INC_PATH,stringToSymbol("SPESTYX"));
  SPP_init(pPreParMap);
  Scn_get_styx(&scn);
  cstream = Stream_file(scn,"SPESTYX",language,".sty");
  Stream_premac_set(cstream, SPP_premac);
  plr     = PLR_get_styx();
  PCfg    = PT_init(plr,cstream);
  if (verbose) fprintf(STDERR,"parsing ...\n");
  PTree   = (styx)PT_PARSE(PCfg,"Source");
  PT_setErrorCnt(PT_synErrorCnt(PCfg));
  PT_quit(PCfg);
  PLR_delTab(plr);
  Stream_free(cstream);
  Scn_free(scn);
  if (verbose) fprintf(STDERR,"done parsing\n");

  if( PT_errorCnt() == 0 )
  {
    styx_Start_0(PTree,&s);
    styxSource_root(s,&id,_,&optcfg);
    if ( strcmp( symbolToString(GLS_Tok_symbol(id)),language) != 0)
      PT_error(id,"language identifier und file name are different");
  
    gloty = MAP_newTyp(primCopy,primFree,PT_equal,PT_hash, primCopy,primFree);
    glo   = MAP_newMap(gloty);
    tkset = HS_CREATE_ADTSET(symbol);

    if (verbose) fprintf(STDERR,"Pass 1\n"); Pass1(s); /* Pass 1: uniqueness */
    if (verbose) fprintf(STDERR,"Pass 2\n"); Pass2(s); /* Pass 2: references */
  }
  if (PT_errorCnt() > 0)
  {
    fprintf(STDERR,"Totally %d errors.\n",PT_errorCnt()); goto EXIT_STYX;
  }

  /* QLX-Uebersetzer */

  if (verbose) fprintf(STDERR,"begin scanner creation\n");

  scn = trans_QlxDfns(s,diagnose);
  if ((Scn_T)scn == NULL) goto EXIT_STYX;
  if (verbose) fprintf(STDERR,"end scanner creation\n");

  if (verbose) fprintf(STDERR,"Pass 3\n"); Pass3(scn,s); /* Pass 3: keyword */
  if (PT_errorCnt() > 0) { Scn_free(scn); goto EXIT_STYX; }
  if (CTX_EVAL(bool,"makeIMG")) Scn_put("BINSTYX",scn);
  if (CTX_EVAL(bool,"makeC")) Scn_C_Modul("GENSTYX",scn);

  /* CFG-Uebersetzer */

  if( styxOptCfg_cfg(optcfg,&dfns) && ! GLS_EMPTY(dfns) )
  {
    PTab = PLR_createTab( (Cfg = trans_Dfns(s,scn)), verbose, diagnose );
    PLR_delCfg(Cfg);

    if( PTab != (PLR_Tab)NULL )
    {
      if( acfg_init(PTab,&aPol,&aNtm,&aPrd,True) )
      {
        genabs = False; genInterface();
        genabs = True;  genInterface();
                        genTypesCom();
        if (CTX_EVAL(bool,"makeIMG")) PLR_putTab("BINSTYX",PTab);
        if (CTX_EVAL(bool,"makeC"))   PLR_printCTab("GENSTYX",PTab);
      }
      acfg_quit(PTab,aPol,aNtm,aPrd);
      PLR_delTab(PTab);
    }
  }
  else
    fprintf(STDERR,"Note: No parser specified.\n");
  if (!CTX_EVAL(bool,"makeIMG") &&
      !CTX_EVAL(bool,"makeINT") &&
      !CTX_EVAL(bool,"makeC"  )  )
    fprintf(STDERR,"Note: No file created.\n");

  Scn_free(scn); /* wg. IgnoreCase an dieser Stelle */
EXIT_STYX:
  if( glo != (MAP(PT_Term, PT_Term))NULL ) MAP_freeMap( glo );
  if( gloty != (MAPTY)NULL )               MAP_freeTyp( gloty );
  if( tkset != (HS_Set)NULL )              HS_DROP_SET(tkset);
  PT_delT(PTree);
  MAP_freeMap(pPreParMap);
  SPP_quit();
  styx_quitSymbols();
  MAP_quit();
  freeSymbols();
  CTX_quit();
  BUG_CORE;
  return 0;
}

